# -*- coding: utf-8 -*-
"""
Compiler stage 4: Code generation
---------------------------------

This module implements the generation of C++ code for the body of each
UFC function from an (optimized) intermediate representation (OIR).
"""

# Copyright (C) 2009-2017 Anders Logg, Martin Sandve Alnæs, Marie E. Rognes,
# Kristian B. Oelgaard, and others
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

from itertools import chain
from ufl import product

# FFC modules
from ffc.log import info, begin, end, debug_code, dstr

# FFC code generation modules
from ffc.representation import pick_representation, ufc_integral_types

import ffc.uflacs.language.cnodes as L
from ffc.uflacs.language.format_lines import format_indented_lines
from ffc.uflacs.backends.ufc.utils import generate_error
from ffc.uflacs.backends.ufc.generators import ufc_integral, ufc_finite_element, ufc_dofmap, ufc_coordinate_mapping, ufc_form

def generate_code(ir, parameters):
    "Generate code from intermediate representation."

    begin("Compiler stage 4: Generating code")

    full_ir = ir

    # FIXME: This has global side effects
    # Set code generation parameters
    # set_float_formatting(parameters["precision"])
    # set_exception_handling(parameters["convert_exceptions_to_warnings"])

    # Extract representations
    ir_finite_elements, ir_dofmaps, ir_coordinate_mappings, ir_integrals, ir_forms = ir

    # Generate code for finite_elements
    info("Generating code for %d finite_element(s)" % len(ir_finite_elements))
    code_finite_elements = [_generate_finite_element_code(ir, parameters)
                            for ir in ir_finite_elements]

    # Generate code for dofmaps
    info("Generating code for %d dofmap(s)" % len(ir_dofmaps))
    code_dofmaps = [_generate_dofmap_code(ir, parameters)
                    for ir in ir_dofmaps]

    # Generate code for coordinate_mappings
    info("Generating code for %d coordinate_mapping(s)" % len(ir_coordinate_mappings))
    code_coordinate_mappings = [_generate_coordinate_mapping_code(ir, parameters)
                                for ir in ir_coordinate_mappings]

    # Generate code for integrals
    info("Generating code for integrals")
    code_integrals = [_generate_integral_code(ir, parameters)
                      for ir in ir_integrals]

    # Generate code for forms
    info("Generating code for forms")
    code_forms = [_generate_form_code(ir, parameters)
                  for ir in ir_forms]

    # Extract additional includes
    includes = _extract_includes(full_ir, code_integrals)

    end()

    return (code_finite_elements, code_dofmaps, code_coordinate_mappings,
            code_integrals, code_forms, includes)


def _extract_includes(full_ir, code_integrals):
    ir_finite_elements, ir_dofmaps, ir_coordinate_mappings, ir_integrals, ir_forms = full_ir

    # Includes added by representations
    includes = set()
    for code in code_integrals:
        includes.update(code["additional_includes_set"])

    # Includes for dependencies in jit mode
    jit = any(full_ir[i][j]["jit"]
              for i in range(len(full_ir))
              for j in range(len(full_ir[i])))
    if jit:
        dep_includes = set()
        for ir in ir_finite_elements:
            dep_includes.update(_finite_element_jit_includes(ir))
        for ir in ir_dofmaps:
            dep_includes.update(_dofmap_jit_includes(ir))
        for ir in ir_coordinate_mappings:
            dep_includes.update(_coordinate_mapping_jit_includes(ir))
        #for ir in ir_integrals:
        #    dep_includes.update(_integral_jit_includes(ir))
        for ir in ir_forms:
            dep_includes.update(_form_jit_includes(ir))
        includes.update(['#include "%s"' % inc for inc in dep_includes])

    return includes


def _finite_element_jit_includes(ir):
    classnames = ir["create_sub_element"]
    postfix = "_finite_element"
    return [classname.rpartition(postfix)[0] + ".h"
            for classname in classnames]


def _dofmap_jit_includes(ir):
    classnames = ir["create_sub_dofmap"]
    postfix = "_dofmap"
    return [classname.rpartition(postfix)[0] + ".h"
            for classname in classnames]


def _coordinate_mapping_jit_includes(ir):
    classnames = [
        ir["coordinate_finite_element_classname"],
        ir["scalar_coordinate_finite_element_classname"]
        ]
    postfix = "_finite_element"
    return [classname.rpartition(postfix)[0] + ".h"
            for classname in classnames]


def _form_jit_includes(ir):
    # Gather all header names for classes that are separately compiled
    # For finite_element and dofmap the module and header name is the prefix,
    # extracted here with .split, and equal for both classes so we skip dofmap here:
    classnames = list(chain(
        ir["create_finite_element"],
        ir["create_coordinate_finite_element"]
        ))
    postfix = "_finite_element"
    includes = [classname.rpartition(postfix)[0] + ".h"
                for classname in classnames]

    classnames = ir["create_coordinate_mapping"]
    postfix = "_coordinate_mapping"
    includes += [classname.rpartition(postfix)[0] + ".h"
                 for classname in classnames]
    return includes


tt_timing_template = """
    // Initialize timing variables
    static const std::size_t _tperiod = 10000;
    static std::size_t _tcount = 0;
    static auto _tsum = std::chrono::nanoseconds::zero();
    static auto _tavg_best = std::chrono::nanoseconds::max();
    static auto _tmin = std::chrono::nanoseconds::max();
    static auto _tmax = std::chrono::nanoseconds::min();

    // Measure single kernel time
    auto _before = std::chrono::high_resolution_clock::now();
    { // Begin original kernel
%s
    } // End original kernel
    // Measure single kernel time
    auto _after = std::chrono::high_resolution_clock::now();

    // Update time stats
    const std::chrono::seconds _s(1);
    auto _tsingle = _after - _before;
    ++_tcount;
    _tsum += _tsingle;
    _tmin = std::min(_tmin, _tsingle);
    _tmax = std::max(_tmax, _tsingle);

    if (_tcount %% _tperiod == 0 || _tsum > _s)
    {
        // Record best average across batches
        std::chrono::nanoseconds _tavg = _tsum / _tcount;
        if (_tavg_best > _tavg)
            _tavg_best = _tavg;

        // Convert to ns
        auto _tot_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(_tsum).count();
        auto _avg_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(_tavg).count();
        auto _min_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(_tmin).count();
        auto _max_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(_tmax).count();
        auto _avg_best_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(_tavg_best).count();

        // Print report
        std::cout << "FFC tt time:"
                  << "  avg_best = " << _avg_best_ns << " ns,"
                  << "  avg = " << _avg_ns << " ns,"
                  << "  min = " << _min_ns << " ns,"
                  << "  max = " << _max_ns << " ns,"
                  << "  tot = " << _tot_ns << " ns,"
                  << "  n = " << _tcount
                  << std::endl;

        // Reset statistics for next batch
        _tcount = 0;
        _tsum = std::chrono::nanoseconds(0);
        _tmin = std::chrono::nanoseconds::max();
        _tmax = std::chrono::nanoseconds::min();
    }
"""


def _generate_tabulate_tensor_comment(ir, parameters):
    "Generate comment for tabulate_tensor."

    r = ir["representation"]
    integrals_metadata = ir["integrals_metadata"]
    integral_metadata = ir["integral_metadata"]

    comment  = [ L.Comment("This function was generated using '%s' representation" % r),
                 L.Comment("with the following integrals metadata:"),
                 L.Comment(""),
                 L.Comment("\n".join(dstr(integrals_metadata).split("\n")[:-1]))
                 ]
    for i, metadata in enumerate(integral_metadata):
        comment += [ L.Comment(""),
                     L.Comment("and the following integral %d metadata:" % i),
                     L.Comment(""),
                     L.Comment("\n".join(dstr(metadata).split("\n")[:-1]))
                   ]

    return format_indented_lines(L.StatementList(comment).cs_format(0), 1)

def _generate_integral_code(ir, parameters):
    "Generate code for integrals from intermediate representation."
    # Select representation
    r = pick_representation(ir["representation"])

    # Generate code
    # TODO: Drop prefix argument and get from ir:
    code = r.generate_integral_code(ir, ir["prefix"], parameters)

    # Hack for benchmarking overhead in assembler with empty
    # tabulate_tensor
    if parameters["generate_dummy_tabulate_tensor"]:
        code["tabulate_tensor"] = ""

    # Wrapping tabulate_tensor in a timing snippet for benchmarking
    if parameters["add_tabulate_tensor_timing"]:
        code["tabulate_tensor"] = tt_timing_template % code["tabulate_tensor"]
        code["additional_includes_set"] = code.get("additional_includes_set", set())
        code["additional_includes_set"].add("#include <chrono>")
        code["additional_includes_set"].add("#include <iostream>")

    # Generate comment
    code["tabulate_tensor_comment"] = _generate_tabulate_tensor_comment(ir, parameters)

    return code

# TODO: Replace the above with this, currently something is not working
def _new_generate_integral_code(ir, parameters):
    "Generate code for integrals from intermediate representation."
    return ufc_integral(ir["integral_type"]).generate_snippets(L, ir, parameters)


def _generate_finite_element_code(ir, parameters):
    "Generate code for finite_element from intermediate representation."
    return ufc_finite_element().generate_snippets(L, ir, parameters)


def _generate_dofmap_code(ir, parameters):
    "Generate code for dofmap from intermediate representation."
    return ufc_dofmap().generate_snippets(L, ir, parameters)


def _generate_coordinate_mapping_code(ir, parameters):
    "Generate code for coordinate_mapping from intermediate representation."
    return ufc_coordinate_mapping().generate_snippets(L, ir, parameters)


def _generate_form_code(ir, parameters):
    "Generate code for coordinate_mapping from intermediate representation."
    return ufc_form().generate_snippets(L, ir, parameters)



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2017 Anders Logg and Martin Sandve Alnæs, Chris Richardson
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.


# Note: Much of the code in this file is a direct translation
# from the old implementation in FFC, although some improvements
# have been made to the generated code.

def jacobian(L, gdim, tdim, element_cellname):
    J = L.Symbol("J")
    coordinate_dofs = L.Symbol("coordinate_dofs")
    code = [L.Comment("Compute Jacobian"),
            L.ArrayDecl("double", J, (gdim*tdim,)),
            L.Call("compute_jacobian_"+element_cellname+"_"+str(gdim)+"d",
                   (J, coordinate_dofs))]
    return code

def inverse_jacobian(L, gdim, tdim, element_cellname):
    K = L.Symbol("K")
    J = L.Symbol("J")
    detJ = L.Symbol("detJ")
    code = [L.Comment("Compute Inverse Jacobian and determinant"),
            L.ArrayDecl("double", K, (gdim*tdim,)),
            L.VariableDecl("double", detJ),
            L.Call("compute_jacobian_inverse_"+element_cellname+"_"+str(gdim)+"d",(K, detJ, J))]
    return code

def orientation(L):
    detJ = L.Symbol("detJ")
    cell_orientation = L.Symbol("cell_orientation")
    code = [L.Comment("Check orientation"),
            L.If(L.EQ(cell_orientation, -1),
                 [L.Throw("std::runtime_error", "cell orientation must be defined (not -1)")]),
            L.Comment("(If cell_orientation == 1 = down, multiply det(J) by -1)"),
            L.ElseIf(L.EQ(cell_orientation, 1),
                     [L.AssignMul(detJ , -1)])]
    return code

def fiat_coordinate_mapping(L, cellname, gdim, ref_coord_symbol="Y"):

    # Code used in evaluatebasis[|derivatives]
    x = L.Symbol("x")
    Y = L.Symbol(ref_coord_symbol)
    coordinate_dofs = L.Symbol("coordinate_dofs")

    if cellname == "interval":
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        if gdim == 1:
            code = [L.Comment("Get coordinates and map to the reference (FIAT) element"),
                    L.ArrayDecl("double", Y, 1, [(2*x[0] - coordinate_dofs[0] - coordinate_dofs[1])/J[0]])]
        elif gdim == 2:
            code = [L.Comment("Get coordinates and map to the reference (FIAT) element"),
                    L.ArrayDecl("double", Y, 1, [2*(L.Sqrt(L.Call("std::pow", (x[0] - coordinate_dofs[0], 2)) + L.Call("std::pow", (x[1] - coordinate_dofs[1], 2))))/detJ - 1.0])]
        elif gdim == 3:
            code = [L.Comment("Get coordinates and map to the reference (FIAT) element"),
                    L.ArrayDecl("double", Y, 1, [2*(L.Sqrt(L.Call("std::pow", (x[0] - coordinate_dofs[0], 2)) + L.Call("std::pow", (x[1] - coordinate_dofs[1], 2)) + L.Call("std::pow", (x[2] - coordinate_dofs[2], 2))))/ detJ - 1.0])]
        else:
            error("Cannot compute interval with gdim: %d" % gdim)
    elif cellname == "triangle":
        if gdim == 2:
            C0 = L.Symbol("C0")
            C1 = L.Symbol("C1")
            J = L.Symbol("J")
            detJ = L.Symbol("detJ")
            code = [L.Comment("Compute constants"),
                    L.VariableDecl("const double", C0, coordinate_dofs[2] + coordinate_dofs[4]),
                    L.VariableDecl("const double", C1, coordinate_dofs[3] + coordinate_dofs[5]),
                    L.Comment("Get coordinates and map to the reference (FIAT) element"),
                    L.ArrayDecl("double", Y, 2, [(J[1]*(C1 - 2.0*x[1]) + J[3]*(2.0*x[0] - C0)) / detJ,
                                                 (J[0]*(2.0*x[1] - C1) + J[2]*(C0 - 2.0*x[0])) / detJ])]
        elif gdim == 3:
            K = L.Symbol("K")
            code = [L.Comment("P_FFC = J^dag (p - b), P_FIAT = 2*P_FFC - (1, 1)"),
                    L.ArrayDecl("double", Y, 2, [2*(K[0]*(x[0] - coordinate_dofs[0])
                                                    + K[1]*(x[1] - coordinate_dofs[1])
                                                    + K[2]*(x[2] - coordinate_dofs[2])) - 1.0,
                                                 2*(K[3]*(x[0] - coordinate_dofs[0])
                                                    + K[4]*(x[1] - coordinate_dofs[1])
                                                    + K[5]*(x[2] - coordinate_dofs[2])) - 1.0])]
        else:
            error("Cannot compute triangle with gdim: %d" % gdim)
    elif cellname == 'tetrahedron' and gdim == 3:
        C0 = L.Symbol("C0")
        C1 = L.Symbol("C1")
        C2 = L.Symbol("C2")
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        d = L.Symbol("d")

        code = [L.Comment("Compute constants"),
                L.VariableDecl("const double", C0, coordinate_dofs[9]  + coordinate_dofs[6] + coordinate_dofs[3] - coordinate_dofs[0]),
                L.VariableDecl("const double", C1, coordinate_dofs[10] + coordinate_dofs[7] + coordinate_dofs[4] - coordinate_dofs[1]),
                L.VariableDecl("const double", C2, coordinate_dofs[11] + coordinate_dofs[8] + coordinate_dofs[5] - coordinate_dofs[2]),
                L.Comment("Compute subdeterminants"),
                L.ArrayDecl("const double", d, 9, [J[4]*J[8] - J[5]*J[7],
                                                   J[5]*J[6] - J[3]*J[8],
                                                   J[3]*J[7] - J[4]*J[6],
                                                   J[2]*J[7] - J[1]*J[8],
                                                   J[0]*J[8] - J[2]*J[6],
                                                   J[1]*J[6] - J[0]*J[7],
                                                   J[1]*J[5] - J[2]*J[4],
                                                   J[2]*J[3] - J[0]*J[5],
                                                   J[0]*J[4] - J[1]*J[3]]),
                L.Comment("Get coordinates and map to the reference (FIAT) element"),
                L.ArrayDecl("double", Y, 3, [(d[0]*(2.0*x[0] - C0) + d[3]*(2.0*x[1] - C1) + d[6]*(2.0*x[2] - C2)) / detJ,
                                             (d[1]*(2.0*x[0] - C0) + d[4]*(2.0*x[1] - C1) + d[7]*(2.0*x[2] - C2)) / detJ,
                                             (d[2]*(2.0*x[0] - C0) + d[5]*(2.0*x[1] - C1) + d[8]*(2.0*x[2] - C2)) / detJ])]
    else:
        error("Cannot compute %s with gdim: %d" % (cellname, gdim))

    return code


def _mapping_transform(L, data, dof_data, values, offset, width=1):

    code = []
    tdim = data["topological_dimension"]
    gdim = data["geometric_dimension"]
    num_components = dof_data["num_components"]
    mapping = dof_data["mapping"]

    # Apply transformation if applicable.
    if mapping == "affine":
        return code

    # Get temporary values before mapping.
    tmp_ref = L.Symbol("tmp_ref")
    code += [L.ArrayDecl("const double", tmp_ref, num_components,
                [values[i*width + offset] for i in range(num_components)])]

    # Define symbols for Jacobian, inverse and determinant
    J = L.Symbol("J")
    J = L.FlattenedArray(J, dims=(gdim, tdim))
    detJ = L.Symbol("detJ")
    K = L.Symbol("K")
    K = L.FlattenedArray(K, dims=(tdim, gdim))

    if mapping == "contravariant piola":
        code += [L.Comment("Using contravariant Piola transform to map values back to the physical element")]
        assert num_components == tdim
        for i in range(gdim):
            inner = sum(J[i, j]*tmp_ref[j] for j in range(tdim))
            code += [L.Assign(values[i*width + offset], inner/detJ)]

    elif mapping == "covariant piola":
        code += [L.Comment("Using covariant Piola transform to map values back to the physical element")]
        assert num_components == tdim
        for i in range(gdim):
            inner = sum(K[j, i]*tmp_ref[j] for j in range(tdim))
            code += [L.Assign(values[i*width + offset], inner)]

    elif mapping == "double covariant piola":
        code += [L.Comment("Using double covariant Piola transform to map values back to the physical element")]
        assert num_components == tdim**2
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim  # noqa: E741
            # g_il = K_ji G_jk K_kl
            inner = sum(K[j, i]*tmp_ref[j*tdim + k]*K[k, l]
                        for j in range(tdim) for k in range(tdim))

            code += [L.Assign(values[p*width + offset], inner)]

    elif mapping == "double contravariant piola":
        code += [L.Comment("Using double contravariant Piola transform to map values back to the physical element")]
        assert num_components == tdim**2
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim  # noqa: E741
            inner = sum(J[i, j]*tmp_ref[j*tdim + k]*J[l, k]
                        for j in range(tdim) for k in range(tdim))

            code += [L.Assign(values[p*width + offset], inner/(detJ*detJ))]
    else:
        error("Unknown mapping: %s" % mapping)
    return code



# -*- coding: utf-8 -*-
"""Work in progress translation of FFC evaluatebasis code to uflacs CNodes format."""

from six import string_types
import numpy

from ffc.log import error
from ffc.uflacs.backends.ufc.utils import generate_error

from ffc.uflacs.backends.ufc.evaluatebasis import generate_expansion_coefficients, generate_compute_basisvalues


# Used for various indices and arrays in this file
index_type = "std::size_t"


def generate_evaluate_reference_basis_derivatives(L, data, parameters):
    # Cutoff for feature to disable generation of this code (consider removing after benchmarking final result)
    if isinstance(data, string_types):
        msg = "evaluate_reference_basis_derivatives: %s" % (data,)
        return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

    # Get some known dimensions
    element_cellname = data["cellname"]
    tdim = data["topological_dimension"]
    max_degree = data["max_degree"]
    reference_value_size = data["reference_value_size"]
    num_dofs = len(data["dofs_data"])

    # Output argument
    reference_values = L.Symbol("reference_values")

    # Input arguments
    order = L.Symbol("order")
    num_points = L.Symbol("num_points")
    X = L.Symbol("X")

    # Loop indices
    ip = L.Symbol("ip") # point
    idof = L.Symbol("i")   # dof
    c = L.Symbol("c")   # component
    r = L.Symbol("r")   # derivative number

    # Define symbol for number of derivatives of given order
    num_derivatives = L.Symbol("num_derivatives")
    reference_values_size = num_points * num_dofs * num_derivatives * reference_value_size

    # FIXME: validate these dimensions
    ref_values = L.FlattenedArray(reference_values,
                                  dims=(num_points, num_dofs, num_derivatives,
                                        reference_value_size))
    # From evaluatebasis.py:
    #ref_values = L.FlattenedArray(reference_values, dims=(num_points, num_dofs, reference_value_size))

    # Initialization (zeroing) and cutoffs outside valid range of orders
    setup_code = [
        # Cutoff to evaluate_basis for order 0
        L.If(L.EQ(order, 0), [
            L.Call("evaluate_reference_basis", (reference_values, num_points, X)),
            L.Return()
            ]),
        # Compute number of derivatives of this order
        L.VariableDecl("const " + index_type, num_derivatives,
                       value=L.Call("std::pow", (tdim, order))),
        # Reset output values to zero
        L.MemZero(reference_values, reference_values_size),
        # Cutoff for higher order than we have
        L.If(L.GT(order, max_degree),
             L.Return()),
        ]

    # If max_degree is zero, we don't need to generate any more code
    if max_degree == 0:
        return setup_code

    # Tabulate dmats tables for all dofs and all derivative directions
    dmats_names, dmats_code = generate_tabulate_dmats(L, data["dofs_data"])

    # Generate code with static tables of expansion coefficients
    tables_code, coefficients_for_dof = generate_expansion_coefficients(L, data["dofs_data"])


    # Generate code to compute tables of basisvalues
    basisvalues_code, basisvalues_for_degree, need_fiat_coordinates = \
        generate_compute_basisvalues(L, data["dofs_data"], element_cellname, tdim, X, ip)

    # Generate all possible combinations of derivatives.
    combinations_code, combinations = _generate_combinations(L, tdim, max_degree, order, num_derivatives)

    # Define symbols for variables populated inside dof switch
    derivatives = L.Symbol("derivatives")
    reference_offset = L.Symbol("reference_offset")
    num_components = L.Symbol("num_components")

    # Get number of components of each basis function (>1 for dofs of piola mapped subelements)
    num_components_values = [dof_data["num_components"] for dof_data in data["dofs_data"]]

    # Offset into parent mixed element to first component of each basis function
    reference_offset_values = [dof_data["reference_offset"] for dof_data in data["dofs_data"]]

    # Max dimensions for the reference derivatives for each dof
    max_num_derivatives = tdim**max_degree
    max_num_components = max(num_components_values)

    # Add constant tables of these numbers
    tables_code += [
        L.ArrayDecl("const " + index_type, reference_offset, num_dofs, values=reference_offset_values),
        L.ArrayDecl("const " + index_type, num_components, num_dofs, values=num_components_values),
    ]

    # Access reference derivatives compactly
    derivs = L.FlattenedArray(derivatives, dims=(num_components[idof], num_derivatives))

    # Create code for all basis values (dofs).
    dof_cases = []
    for i_dof, dof_data in enumerate(data["dofs_data"]):

        embedded_degree = dof_data["embedded_degree"]
        basisvalues = basisvalues_for_degree[embedded_degree]

        shape_dmats = numpy.shape(dof_data["dmats"][0])
        if shape_dmats[0] != shape_dmats[1]:
            error("Something is wrong with the dmats:\n%s" % str(dof_data["dmats"]))

        aux = L.Symbol("aux")
        dmats = L.Symbol("dmats")
        dmats_old = L.Symbol("dmats_old")
        dmats_name = dmats_names[i_dof]

        # Create dmats matrix by multiplication
        comb = L.Symbol("comb")
        s = L.Symbol("s")
        t = L.Symbol("t")
        u = L.Symbol("u")
        tu = L.Symbol("tu")
        aux_computation_code = [ L.ArrayDecl("double", aux, shape_dmats[0], values=0),
                            L.Comment("Declare derivative matrix (of polynomial basis)."),
                            L.ArrayDecl("double", dmats, shape_dmats, values=0),
                            L.Comment("Initialize dmats."),
                            L.VariableDecl(index_type, comb, combinations[r, 0]),
                            L.MemCopy(L.AddressOf(dmats_name[comb][0][0]),  L.AddressOf(dmats[0][0]), shape_dmats[0]*shape_dmats[1]),
                            L.Comment("Looping derivative order to generate dmats."),
                            L.ForRange(s, 1, order, index_type=index_type, body=[
                                L.Comment("Store previous dmats matrix."),
                                L.ArrayDecl("double", dmats_old, shape_dmats),
                                L.MemCopy(L.AddressOf(dmats[0][0]),  L.AddressOf(dmats_old[0][0]), shape_dmats[0]*shape_dmats[1]),
                                L.Comment("Resetting dmats."),
                                L.MemZero(L.AddressOf(dmats[0][0]), shape_dmats[0]*shape_dmats[1]),
                                L.Comment("Update dmats using an inner product."),
                                L.Assign(comb, combinations[r, s]),
                                L.ForRange(t, 0, shape_dmats[0], index_type=index_type, body=
                                   L.ForRange(u, 0, shape_dmats[1], index_type=index_type, body=
                                      L.ForRange(tu, 0, shape_dmats[0], index_type=index_type, body=
                                         L.AssignAdd(dmats[t, u], dmats_name[comb, t, tu] * dmats_old[tu, u]))))]),
                            L.ForRange(s, 0, shape_dmats[0], index_type=index_type, body=
                               L.ForRange(t, 0, shape_dmats[1], index_type=index_type, body=
                                          L.AssignAdd(aux[s], dmats[s, t] * basisvalues[t])))
        ]

        # Unrolled loop over components of basis function
        n = dof_data["num_components"]
        compute_ref_derivs_code = [L.Assign(derivs[cc][r], 0.0) for cc in range(n)]

        compute_ref_derivs_code += [L.ForRange(s, 0, shape_dmats[0], index_type=index_type, body=
                                               [L.AssignAdd(derivs[cc][r], coefficients_for_dof[i_dof][cc][s] * aux[s]) for cc in range(n)])]

        embedded_degree = dof_data["embedded_degree"]
        basisvalues = basisvalues_for_degree[embedded_degree]

        # Compute the derivatives of the basisfunctions on the reference (FIAT) element,
        # as the dot product of the new coefficients and basisvalues.

        case_code =  [L.Comment("Compute reference derivatives for dof %d." % i_dof),
                      # Accumulate sum_s coefficients[s] * aux[s]
                      L.ForRange(r, 0, num_derivatives, index_type=index_type, body=[
                          aux_computation_code,
                          compute_ref_derivs_code
                      ])]

        dof_cases.append((i_dof, case_code))

    # Loop over all dofs, entering a different switch case in each iteration.
    # This is a legacy from the previous implementation where the loop
    # was in a separate function and all the setup above was also repeated
    # in a call for each dof.
    # TODO: Further refactoring is needed to improve on this situation,
    # but at least it's better than before. There's probably more code and
    # computations that can be shared between dofs, and this would probably
    # be easier to fix if mixed elements were handled separately!
    dof_loop_code = [
        L.Comment("Loop over all dofs"),
        L.ForRange(idof, 0, num_dofs, index_type=index_type, body=[
            L.ArrayDecl("double", derivatives, max_num_components * max_num_derivatives, 0.0),
            L.Switch(idof, dof_cases),
            L.ForRange(r, 0, num_derivatives, index_type=index_type, body= [
                L.ForRange(c, 0, num_components[idof], index_type=index_type, body=[
                    L.Assign(ref_values[ip][idof][r][reference_offset[idof] + c], derivs[c][r]),  # FIXME: validate ref_values dims
                ]),
            ])
        ]),
    ]

    # Define loop over points
    final_loop_code = [
        L.ForRange(ip, 0, num_points, index_type=index_type, body=
            basisvalues_code
            + dof_loop_code
            )
        ]

    # Stitch it all together
    code = (
        setup_code
        + dmats_code
        + tables_code
        + combinations_code
        + final_loop_code
        )

    return code


def generate_tabulate_dmats(L, dofs_data):
    "Tabulate the derivatives of the polynomial base"

    alignas = 32

    # Emit code for the dmats we've actually used
    dmats_code = [L.Comment("Tables of derivatives of the polynomial base (transpose).")]

    dmats_names = []

    all_matrices = []

    for idof, dof_data in enumerate(dofs_data):
        # Get derivative matrices (coefficients) of basis functions, computed by FIAT at compile time.
        derivative_matrices = dof_data["dmats"]
        num_mats = len(derivative_matrices)
        num_members = dof_data["num_expansion_members"]

        # Generate tables for each spatial direction.
        matrix = numpy.zeros((num_mats, num_members, num_members))
        for i, dmat in enumerate(derivative_matrices):
            # Extract derivatives for current direction
            # (take transpose, FIAT_NEW PolynomialSet.tabulate()).
            matrix[i,...] = numpy.transpose(dmat)

        # TODO: Use precision from parameters here
        from ffc.uflacs.elementtables import clamp_table_small_numbers
        matrix = clamp_table_small_numbers(matrix)

        # O(n^2) matrix matching...
        name = None
        for oldname, oldmatrix in all_matrices:
            if matrix.shape == oldmatrix.shape and numpy.allclose(matrix, oldmatrix):
                name = oldname
                break

        if name is None:
            # Define variable name for coefficients for this dof
            name = L.Symbol("dmats%d" % (idof,))
            all_matrices.append((name, matrix))

            # Declare new dmats table with unique values
            decl = L.ArrayDecl("static const double", name, (num_mats, num_members, num_members),
                               values=matrix, alignas=alignas)
            dmats_code.append(decl)

        # Append name for each dof
        dmats_names.append(name)

    return dmats_names, dmats_code


def _generate_combinations(L, tdim, max_degree, order, num_derivatives, suffix=""):
    max_num_derivatives = tdim**max_degree
    combinations = L.Symbol("combinations" + suffix)

    # This precomputes the combinations for each order and stores in code as table
    # Python equivalent precomputed for each valid order:
    combinations_shape = (max_degree, max_num_derivatives, max_degree)
    all_combinations = numpy.zeros(combinations_shape, dtype=int)
    for q in range(1, max_degree+1):
        for row in range(1, max_num_derivatives):
            for num in range(0, row):
                for col in range(q-1, -1, -1):
                    if all_combinations[q-1][row][col] > tdim - 2:
                        all_combinations[q-1][row][col] = 0
                    else:
                        all_combinations[q-1][row][col] += 1
                        break
    code = [
        L.Comment("Precomputed combinations"),
        L.ArrayDecl("const " + index_type, combinations, combinations_shape, values=all_combinations),
        ]
    # Select the right order for further access
    combinations = combinations[order-1]

    return code, combinations



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2017 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

import re
from six import string_types

from ffc.log import error, warning
import ffc.backends.ufc

from ffc.uflacs.language.format_lines import format_indented_lines
from ffc.uflacs.backends.ufc.templates import *

#__all__ = (["ufc_form", "ufc_dofmap", "ufc_finite_element", "ufc_integral"]
#           + ["ufc_%s_integral" % integral_type for integral_type in integral_types])


# These are all the integral types directly supported in ufc
from ffc.representation import ufc_integral_types


# These are the method names in ufc::form that are specialized for each integral type
integral_name_templates = (
    "max_%s_subdomain_id",
    "has_%s_integrals",
    "create_%s_integral",
    "create_default_%s_integral",
    )


class ufc_generator(object):
    """Common functionality for code generators producing ufc classes.

    The generate function is the driver for generating code for a class.
    It automatically extracts each template keyword %(foo)s and calls
    self.foo(...) to define the code snippet for that keyword.

    The standard arguments to every self.foo(...) function are:
    - L (the language module reference)
    - ir (the full ir dict)
    - parameters (the parameters dict, can be omitted)
    If the second argument is not named "ir", it must be
    a valid key in ir, and the value of ir[keyname] is
    passed instead of the full ir dict. Invalid keynames
    result in attempts at informative errors, meaning errors
    will often be caught early when making changes.
    """
    def __init__(self, basename):
        ufc_templates = ffc.backends.ufc.templates
        self._header_template = ufc_templates[basename + "_header"]
        self._implementation_template = ufc_templates[basename + "_implementation"]
        self._combined_template = ufc_templates[basename + "_combined"]
        self._jit_header_template = ufc_templates[basename + "_jit_header"]
        self._jit_implementation_template = ufc_templates[basename + "_jit_implementation"]

        r = re.compile(r"%\(([a-zA-Z0-9_]*)\)")
        self._header_keywords = set(r.findall(self._header_template))
        self._implementation_keywords = set(r.findall(self._implementation_template))
        self._combined_keywords = set(r.findall(self._combined_template))

        self._keywords = sorted(self._header_keywords | self._implementation_keywords)

        # Do some ufc interface template checking, to catch bugs
        # early when we change the ufc interface templates
        if set(self._keywords) != set(self._combined_keywords):
            a = set(self._header_keywords) - set(self._combined_keywords)
            b = set(self._implementation_keywords) - set(self._combined_keywords)
            c = set(self._combined_keywords) - set(self._keywords)
            msg = "Templates do not have matching sets of keywords:"
            if a:
                msg += "\n  Header template keywords '%s' are not in the combined template." % (sorted(a),)
            if b:
                msg += "\n  Implementation template keywords '%s' are not in the combined template." % (sorted(b),)
            if c:
                msg += "\n  Combined template keywords '%s' are not in the header or implementation templates." % (sorted(c),)
            error(msg)

    def generate_snippets(self, L, ir, parameters):
        "Generate code snippets for each keyword found in templates."
        snippets = {}
        for kw in self._keywords:
            handlerstr = "%s.%s" % (self.__class__.__name__, kw)

            # Check that attribute self.<keyword> is available
            if not hasattr(self, kw):
                error("Missing handler %s." % (handlerstr,))

            # Call self.<keyword>(*args) to get value to insert in snippets
            method = getattr(self, kw)
            vn = method.__code__.co_varnames[:method.__code__.co_argcount]
            file_line = "%s:%s" % (method.__code__.co_filename, method.__code__.co_firstlineno)

            #if handlerstr == "ufc_dofmap.create":
            #    import ipdb; ipdb.set_trace()

            # Always pass L
            assert vn[:2] == ("self", "L")
            vn = vn[2:]
            args = (L,)

            # Either pass full ir or extract ir value with keyword given by argument name
            if vn[0] == "ir":
                args += (ir,)
            elif vn[0] in ir:
                args += (ir[vn[0]],)
            else:
                error("Cannot find key '%s' in ir, argument to %s at %s." % (vn[0], handlerstr, file_line))
            vn = vn[1:]

            # Optionally pass parameters
            if vn == ("parameters",):
                args += (parameters,)
            elif vn:
                error("Invalid argument names %s to %s at %s." % (vn, handlerstr, file_line))

            # Call handler
            value = method(*args)


            if isinstance(value, list):
                value = L.StatementList(value)

            # Indent body and format to str
            if isinstance(value, L.CStatement):
                value = L.Indented(value.cs_format(precision=parameters["precision"]))
                value = format_indented_lines(value)
            elif not isinstance(value, string_types):
                error("Expecting code or string, not %s, returned from handler %s at %s." % (type(value), handlerstr, file_line))

            # Store formatted code in snippets dict
            snippets[kw] = value

        # Error checking (can detect some bugs early when changing the ufc interface)
        # Get all attributes of subclass class (skip "_foo")
        attrs = set(name for name in dir(self) if not (name.startswith("_") or name.startswith("generate")))
        # Get all attributes of this base class (skip "_foo" and "generate*")
        base_attrs = set(name for name in dir(ufc_generator) if not (name.startswith("_") or name.startswith("generate")))
        # The template keywords should not contain any names not among the class attributes
        missing = set(self._keywords) - attrs
        if missing:
            warning("*** Missing generator functions:\n%s" % ('\n'.join(map(str, sorted(missing))),))
        # The class attributes should not contain any names not among the template keywords
        # (this is strict, a useful check when changing ufc, but can be dropped)
        unused = attrs - set(self._keywords) - base_attrs
        if unused:
            warning("*** Unused generator functions:\n%s" % ('\n'.join(map(str, sorted(unused))),))

        # Return snippets, a dict of code strings
        return snippets

    def generate(self, L, ir, parameters=None, snippets=None, jit=False):
        "Return composition of templates with generated snippets."
        if snippets is None:
            snippets = self.generate_snippets(L, ir, parameters)
        if jit:
            ht = self._jit_header_template
            it = self._jit_implementation_template
        else:
            ht = self._header_template
            it = self._implementation_template
        h = ht % snippets
        cpp = it % snippets
        return h, cpp

    def classname(self, L, ir):
        "Return classname."
        return ir["classname"]

    def members(self, L, ir):
        "Return empty string. Override in classes that need members."
        if ir.get("members"):
            error("Missing generator function.")
        return ""

    def constructor(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("constructor"):
            error("Missing generator function.")
        return L.NoOp()

    def constructor_arguments(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("constructor_arguments"):
            error("Missing generator function.")
        return ""

    def initializer_list(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("initializer_list"):
            error("Missing generator function.")
        return ""

    def destructor(self, L, ir):
        "Return empty string. Override in classes that need destructor."
        if ir.get("destructor"):
            error("Missing generator function.")
        return L.NoOp()

    def signature(self, L, ir):
        "Default implementation of returning signature string fetched from ir."
        sig = ir["signature"]
        return L.Return(L.LiteralString(sig))

    def create(self, L, ir):
        "Default implementation of creating a new object of the same type."
        classname = ir["classname"]
        return L.Return(L.New(classname))



# -*- coding: utf-8 -*-

class QuadratureRepresentationDeprecationWarning(DeprecationWarning):
    """Warning about deprecation of quadrature representation"""
    def __init__(self):
        msg = """
*** ===================================================== ***
*** FFC: quadrature representation is deprecated! It will ***
*** likely be removed in 2018.1.0 release. Use uflacs     ***
*** representation instead.                               ***
*** ===================================================== ***"""
        super(QuadratureRepresentationDeprecationWarning, self).__init__(msg)


# Be very annoying - never ignore the warning
import warnings
warnings.simplefilter('always', QuadratureRepresentationDeprecationWarning)
del warnings


def issue_deprecation_warning():
    """Issue warning about quadrature deprecation"""
    import warnings
    warnings.warn(QuadratureRepresentationDeprecationWarning(), stacklevel=2)



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2017 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.


# Note: Much of the code in this file is a direct translation
# from the old implementation in FFC, although some improvements
# have been made to the generated code.


from collections import defaultdict
import numpy
from six import string_types

from ufl import product
from ffc.uflacs.backends.ufc.generator import ufc_generator
from ffc.uflacs.backends.ufc.utils import generate_return_new_switch, generate_return_int_switch, generate_error

from ffc.uflacs.elementtables import clamp_table_small_numbers
from ffc.uflacs.backends.ufc.evaluatebasis import generate_evaluate_reference_basis
from ffc.uflacs.backends.ufc.evaluatebasis import _generate_compute_basisvalues, tabulate_coefficients
from ffc.uflacs.backends.ufc.evaluatebasisderivatives import generate_evaluate_basis_derivatives_all
from ffc.uflacs.backends.ufc.evaluatebasisderivatives import generate_evaluate_basis_derivatives
from ffc.uflacs.backends.ufc.evalderivs import generate_evaluate_reference_basis_derivatives
from ffc.uflacs.backends.ufc.evalderivs import _generate_combinations
from ffc.uflacs.backends.ufc.evaluatedof import generate_evaluate_dof, generate_evaluate_dofs, reference_to_physical_map

from ffc.uflacs.backends.ufc.jacobian import jacobian, inverse_jacobian, orientation, fiat_coordinate_mapping, _mapping_transform

index_type = "std::size_t"

def compute_basis_values(L, data, dof_data):
    basisvalues = L.Symbol("basisvalues")
    Y = L.Symbol("Y")
    element_cellname = data["cellname"]
    embedded_degree = dof_data["embedded_degree"]
    num_members = dof_data["num_expansion_members"]
    return _generate_compute_basisvalues(L, basisvalues, Y, element_cellname,
                                         embedded_degree, num_members)


def compute_values(L, data, dof_data):
    """This function computes the value of the basisfunction as the dot product
    of the coefficients and basisvalues."""

    # Initialise return code.
    code = [L.Comment("Compute value(s)")]

    # Get dof data.
    num_components = dof_data["num_components"]
    physical_offset = dof_data["physical_offset"]

    basisvalues = L.Symbol("basisvalues")
    values = L.Symbol("values")
    r = L.Symbol("r")
    lines = []
    if data["reference_value_size"] != 1:
        # Loop number of components.
        for i in range(num_components):
            coefficients = L.Symbol("coefficients%d" % i)
            lines += [L.AssignAdd(values[i + physical_offset], coefficients[r]*basisvalues[r])]
    else:
        coefficients = L.Symbol("coefficients0")
        lines = [L.AssignAdd(L.Dereference(values), coefficients[r]*basisvalues[r])]

    # Get number of members of the expansion set and generate loop.
    num_mem = dof_data["num_expansion_members"]
    code += [L.ForRange(r, 0, num_mem, index_type=index_type, body=lines)]

    code += _mapping_transform(L, data, dof_data, values, physical_offset)

    return code

def generate_element_mapping(mapping, i, num_reference_components, tdim, gdim, J, detJ, K):
    # Select transformation to apply
    if mapping == "affine":
        assert num_reference_components == 1
        num_physical_components = 1
        M_scale = 1
        M_row = [1]  # M_row[0] == 1
    elif mapping == "contravariant piola":
        assert num_reference_components == tdim
        num_physical_components = gdim
        M_scale = 1.0 / detJ
        M_row = [J[i, jj] for jj in range(tdim)]
    elif mapping == "covariant piola":
        assert num_reference_components == tdim
        num_physical_components = gdim
        M_scale = 1.0
        M_row = [K[jj, i] for jj in range(tdim)]
    elif mapping == "double covariant piola":
        assert num_reference_components == tdim**2
        num_physical_components = gdim**2
        # g_il = K_ji G_jk K_kl = K_ji K_kl G_jk
        i0 = i // tdim  # i in the line above
        i1 = i % tdim   # l ...
        M_scale = 1.0
        M_row = [K[jj,i0]*K[kk,i1] for jj in range(tdim) for kk in range(tdim)]
    elif mapping == "double contravariant piola":
        assert num_reference_components == tdim**2
        num_physical_components = gdim**2
        # g_il = (det J)^(-2) Jij G_jk Jlk = (det J)^(-2) Jij Jlk G_jk
        i0 = i // tdim  # i in the line above
        i1 = i % tdim   # l ...
        M_scale = 1.0 / (detJ*detJ)
        M_row = [J[i0,jj]*J[i1,kk] for jj in range(tdim) for kk in range(tdim)]
    else:
        error("Unknown mapping: %s" % mapping)
    return M_scale, M_row, num_physical_components


class ufc_finite_element(ufc_generator):
    "Each function maps to a keyword in the template. See documentation of ufc_generator."
    def __init__(self):
        ufc_generator.__init__(self, "finite_element")

    def cell_shape(self, L, cell_shape):
        return L.Return(L.Symbol("ufc::shape::" + cell_shape))

    def topological_dimension(self, L, topological_dimension):
        return L.Return(topological_dimension)

    def geometric_dimension(self, L, geometric_dimension):
        return L.Return(geometric_dimension)

    def space_dimension(self, L, space_dimension):
        return L.Return(space_dimension)

    def value_rank(self, L, value_shape):
        return L.Return(len(value_shape))

    def value_dimension(self, L, value_shape):
        return generate_return_int_switch(L, "i", value_shape, 1)

    def value_size(self, L, value_shape):
        return L.Return(product(value_shape))

    def reference_value_rank(self, L, reference_value_shape):
        return L.Return(len(reference_value_shape))

    def reference_value_dimension(self, L, reference_value_shape):
        return generate_return_int_switch(L, "i", reference_value_shape, 1)

    def reference_value_size(self, L, reference_value_shape):
        return L.Return(product(reference_value_shape))

    def degree(self, L, degree):
        return L.Return(degree)

    def family(self, L, family):
        return L.Return(L.LiteralString(family))

    def num_sub_elements(self, L, num_sub_elements):
        return L.Return(num_sub_elements)

    def create_sub_element(self, L, ir):
        classnames = ir["create_sub_element"]
        return generate_return_new_switch(L, "i", classnames, factory=ir["jit"])

    def evaluate_basis(self, L, ir, parameters):
        data = ir["evaluate_basis"]

        # FIXME: does this make sense?
        if not data:
            msg = "evaluate_basis is not defined for this element"
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        # Handle unsupported elements.
        if isinstance(data, string_types):
            msg = "evaluate_basis: %s" % data
            return [generate_error(L, msg, parameters["convert_exceptions_to_warnings"])]

        # Get the element cell name and geometric dimension.
        element_cellname = data["cellname"]
        gdim = data["geometric_dimension"]
        tdim = data["topological_dimension"]

        # Generate run time code to evaluate an element basisfunction
        # at an arbitrary point. The value(s) of the basisfunction
        # is/are computed as in FIAT as the dot product of the
        # coefficients (computed at compile time) and basisvalues
        # which are dependent on the coordinate and thus have to be
        # computed at run time.

        # The function should work for all elements supported by FIAT,
        # but it remains untested for tensor valued elements.

        # Get code snippets for Jacobian, Inverse of Jacobian and
        # mapping of coordinates from physical element to the FIAT
        # reference element.

        cm = L.Symbol("cm")
        X = L.Symbol("X")
        code = [L.ArrayDecl("double", X, (tdim), values=0)]

        J = L.Symbol("J")
        code += [L.ArrayDecl("double", J, (gdim*tdim,))]

        detJ = L.Symbol("detJ")
        code += [L.VariableDecl("double", detJ)]

        K = L.Symbol("K")
        code += [L.ArrayDecl("double", K, (gdim*tdim,))]

        x = L.Symbol("x")
        coordinate_dofs = L.Symbol("coordinate_dofs")
        cell_orientation = L.Symbol("cell_orientation")

        no_cm_code = [  L.Call("compute_jacobian_"+element_cellname+"_"+str(gdim)+"d",
                               (J, coordinate_dofs)),
                        L.Call("compute_jacobian_inverse_"+element_cellname+"_"+str(gdim)+"d",
                               (K, detJ, J))]

        if data["needs_oriented"] and tdim != gdim:
            no_cm_code += orientation(L)

        if any((d["embedded_degree"] > 0) for d in data["dofs_data"]):
            k = L.Symbol("k")
            Y = L.Symbol("Y")
            no_cm_code += fiat_coordinate_mapping(L, element_cellname, gdim)
            if element_cellname in ('interval', 'triangle', 'tetrahedron'):
                no_cm_code += [L.Comment("Map to FFC reference coordinate"),
                               L.ForRange(k, 0, tdim, index_type=index_type, body=[L.Assign(X[k], (Y[k] + 1.0)/2.0)])]
            else:
                no_cm_code += [L.ForRange(k, 0, tdim, index_type=index_type, body=[L.Assign(X[k], Y[k])])]

        code += [L.If(cm, L.Call("cm->compute_reference_geometry",
                                (X, J, L.AddressOf(detJ), K, 1, x, coordinate_dofs, cell_orientation))),
                 L.Else(no_cm_code)]

        reference_value_size = data["reference_value_size"]
        num_dofs = len(data["dofs_data"])
        ref_values = L.Symbol("ref_values")
        code += [L.Comment("Evaluate basis on reference element"),
                 L.ArrayDecl("double", ref_values, num_dofs*reference_value_size),
                 L.Call("evaluate_reference_basis",(ref_values, 1, X))]

        physical_value_size = data["physical_value_size"]
        physical_values = L.Symbol("physical_values")
        i = L.Symbol("i")
        k = L.Symbol("k")
        values = L.Symbol("values")
        code += [L.Comment("Push forward"),
                 L.ArrayDecl("double", physical_values, num_dofs*physical_value_size),
                 L.Call("transform_reference_basis_derivatives",(physical_values, 0, 1,
                                                                 ref_values, X, J, L.AddressOf(detJ), K,
                                                                 cell_orientation)),
                 L.ForRange(k, 0, physical_value_size, index_type=index_type, body=[
                     L.Assign(values[k], physical_values[physical_value_size*i + k])
                 ])]

        return code

    def evaluate_basis_all(self, L, ir, parameters):
        data=ir["evaluate_basis"]

        # Handle unsupported elements.
        if isinstance(data, string_types):
            msg = "evaluate_basis_all: %s" % data
            return [generate_error(L, msg, parameters["convert_exceptions_to_warnings"])]

        physical_value_size = data["physical_value_size"]
        space_dimension = data["space_dimension"]

        x = L.Symbol("x")
        coordinate_dofs = L.Symbol("coordinate_dofs")
        cell_orientation = L.Symbol("cell_orientation")
        values = L.Symbol("values")

        # Special case where space dimension is one (constant elements).
        if space_dimension == 1:
            code = [L.Comment("Element is constant, calling evaluate_basis."),
                    L.Call("evaluate_basis",
                           (0, values, x, coordinate_dofs, cell_orientation))]
            return code

        r = L.Symbol("r")
        dof_values = L.Symbol("dof_values")
        if physical_value_size == 1:
            code = [ L.Comment("Helper variable to hold value of a single dof."),
                     L.VariableDecl("double", dof_values, 0.0),
                     L.Comment("Loop dofs and call evaluate_basis"),
                     L.ForRange(r, 0, space_dimension, index_type=index_type,
                                body=[L.Call("evaluate_basis",
                                             (r, L.AddressOf(dof_values), x,
                                              coordinate_dofs, cell_orientation)),
                                      L.Assign(values[r], dof_values)]
                               )
                   ]
        else:
            s = L.Symbol("s")
            code = [L.Comment("Helper variable to hold values of a single dof."),
                    L.ArrayDecl("double", dof_values, physical_value_size, 0.0),
                    L.Comment("Loop dofs and call evaluate_basis"),
                    L.ForRange(r, 0, space_dimension, index_type=index_type,
                               body=[L.Call("evaluate_basis",
                                             (r, dof_values, x,
                                              coordinate_dofs, cell_orientation)),
                                     L.ForRange(s, 0, physical_value_size,
                                                index_type=index_type,
                                                body=[L.Assign(values[r*physical_value_size+s], dof_values[s])])
                                    ]
                              )
                   ]

        return code

    def evaluate_basis_derivatives(self, L, ir, parameters):
        # FIXME: Get rid of this
        return generate_evaluate_basis_derivatives(L, ir["evaluate_basis"])

    def evaluate_basis_derivatives_all(self, L, ir, parameters):
        return generate_evaluate_basis_derivatives_all(L, ir["evaluate_basis"])

        """
        // Legacy version:
        evaluate_basis_derivatives_all(std::size_t n,
                                       double * values,
                                       const double * x,
                                       const double * coordinate_dofs,
                                       int cell_orientation)
        // Suggestion for new version:
        new_evaluate_basis_derivatives(double * values,
                                       std::size_t order,
                                       std::size_t num_points,
                                       const double * x,
                                       const double * coordinate_dofs,
                                       int cell_orientation,
                                       const ufc::coordinate_mapping * cm)
        """

        # TODO: This is a refactoring step to allow rewriting code
        # generation to use coordinate_mapping in one stage, before
        # making it available as an argument from dolfin in the next stage.
        affine_coordinate_mapping_classname = ir["affine_coordinate_mapping_classname"]

        # Output arguments:
        values = L.Symbol("values")

        # Input arguments:
        #order = L.Symbol("order")
        order = L.Symbol("n")
        x = L.Symbol("x")
        coordinate_dofs = L.Symbol("coordinate_dofs")
        cell_orientation = L.Symbol("cell_orientation")

        # Internal variables:
        #num_points = L.Symbol("num_points")
        num_points = 1  # Always 1 in legacy API
        reference_values = L.Symbol("reference_values")
        X = L.Symbol("X")
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        K = L.Symbol("K")
        ip = L.Symbol("ip")

        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]

        code = [
            # Create local affine coordinate mapping object
            # TODO: Get this as input instead to support non-affine
            L.VariableDecl(affine_coordinate_mapping_classname, "cm"),
            L.ForRange(ip, 0, num_points, index_type=index_type, body=[
                L.ArrayDecl("double", X, (tdim,)),
                L.ArrayDecl("double", J, (gdim*tdim,)),
                L.ArrayDecl("double", detJ, (1,)),
                L.ArrayDecl("double", K, (tdim*gdim,)),
                L.Call("cm.compute_reference_geometry",
                       (X, J, detJ, K, num_points, x, coordinate_dofs, cell_orientation)),
                L.Call("evaluate_reference_basis_derivatives",
                       (reference_values, order, num_points, X)),
                L.Call("transform_reference_basis_derivatives",
                       (values, order, num_points, reference_values, X, J, detJ, K, cell_orientation)),
            ])
        ]
        return code

    def evaluate_dof(self, L, ir, parameters):
        return generate_evaluate_dof(L, ir["evaluate_dof"])


    def evaluate_dofs(self, L, ir, parameters):
        """Generate code for evaluate_dofs."""
        """
        - evaluate_dof needs to be split into invert_mapping + evaluate_dof or similar?

          f = M fhat;  nu(f) = nu(M fhat) = nuhat(M^-1 f) = sum_i w_i M^-1 f(x_i)

          // Get fixed set of points on reference element
          num_points = element->num_dof_evaluation_points();
          double X[num_points*tdim];
          element->tabulate_dof_evaluation_points(X);

          // Compute geometry in these points
          domain->compute_geometry(reference_points, num_point, X, J, detJ, K, coordinate_dofs, cell_orientation);

          // Computed by dolfin
          for ip
            fvalues[ip][:] = f.evaluate(point[ip])[:];

          // Finally: nu_j(f) = sum_component sum_ip weights[j][ip][component] fvalues[ip][component]
          element->evaluate_dofs(fdofs, fvalues, J, detJ, K)
        """
        # FIXME: translate into reference version
        return generate_evaluate_dofs(L, ir["evaluate_dof"])

    def interpolate_vertex_values(self, L, ir, parameters):
        irdata = ir["interpolate_vertex_values"]

        # Raise error if interpolate_vertex_values is ill-defined
        if not irdata:
            msg = "interpolate_vertex_values is not defined for this element"
            return [generate_error(L, msg, parameters["convert_exceptions_to_warnings"])]

        # Handle unsupported elements.
        if isinstance(irdata, string_types):
            msg = "interpolate_vertex_values: %s" % irdata
            return [generate_error(L, msg, parameters["convert_exceptions_to_warnings"])]

        # Add code for Jacobian if necessary
        code = []
        gdim = irdata["geometric_dimension"]
        tdim = irdata["topological_dimension"]
        cell_shape = ir["cell_shape"]
        if irdata["needs_jacobian"]:
            code += jacobian(L, gdim, tdim, cell_shape)
            code += inverse_jacobian(L, gdim, tdim, cell_shape)
            if irdata["needs_oriented"] and tdim != gdim:
                code += orientation(L)

        # Compute total value dimension for (mixed) element
        total_dim = irdata["physical_value_size"]

        # Generate code for each element
        value_offset = 0
        space_offset = 0
        for data in irdata["element_data"]:
            # Add vertex interpolation for this element
            code += [L.Comment("Evaluate function and change variables")]

            # Extract vertex values for all basis functions
            vertex_values = data["basis_values"]
            value_size = data["physical_value_size"]
            space_dim = data["space_dim"]
            mapping = data["mapping"]

            J = L.Symbol("J")
            J = L.FlattenedArray(J, dims=(gdim, tdim))
            detJ = L.Symbol("detJ")
            K = L.Symbol("K")
            K = L.FlattenedArray(K, dims=(tdim, gdim))

            # Create code for each value dimension:
            for k in range(value_size):
                # Create code for each vertex x_j
                for (j, values_at_vertex) in enumerate(vertex_values):

                    if value_size == 1:
                        values_at_vertex = [values_at_vertex]

                    values = clamp_table_small_numbers(values_at_vertex)

                    # Map basis functions using appropriate mapping
                    # FIXME: sort out all non-affine mappings and make into a function
                    # components = change_of_variables(values_at_vertex, k)

                    w = []
                    if mapping == 'affine':
                        w = values[k]
                    elif mapping == 'contravariant piola':
                        for index in range(space_dim):
                            w += [sum(J[k, p]*values[p][index]
                                      for p in range(tdim))/detJ]
                    elif mapping == 'covariant piola':
                        for index in range(space_dim):
                            w += [sum(K[p, k]*values[p][index]
                                      for p in range(tdim))]
                    elif mapping == 'double covariant piola':
                        for index in range(space_dim):
                            w += [sum(K[p, k//tdim]*values[p][q][index]*K[q, k % tdim]
                                      for q in range(tdim) for p in range(tdim))]
                    elif mapping == 'double contravariant piola':
                        for index in range(space_dim):
                            w += [sum(J[k//tdim, p]*values[p][q][index]*J[k % tdim, q]
                                      for q in range(tdim) for p in range(tdim))/(detJ*detJ)]
                    else:
                        error("Unknown mapping: %s" % mapping)

                    # Contract coefficients and basis functions
                    dof_values = L.Symbol("dof_values")
                    dof_list = [dof_values[i + space_offset] for i in range(space_dim)]
                    value = sum(p*q for (p, q) in zip(dof_list, w))

                    # Assign value to correct vertex
                    index = j * total_dim + (k + value_offset)
                    v_values = L.Symbol("vertex_values")
                    code += [L.Assign(v_values[index], value)]

            # Update offsets for value- and space dimension
            value_offset += data["physical_value_size"]
            space_offset += data["space_dim"]

        return code

    def tabulate_dof_coordinates(self, L, ir, parameters):
        ir = ir["tabulate_dof_coordinates"]

        # Raise error if tabulate_dof_coordinates is ill-defined
        if not ir:
            msg = "tabulate_dof_coordinates is not defined for this element"
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        # Extract coordinates and cell dimension
        gdim = ir["gdim"]
        tdim = ir["tdim"]
        points = ir["points"]

        # Extract cellshape
        cell_shape = ir["cell_shape"]

        # Output argument
        dof_coordinates = L.FlattenedArray(L.Symbol("dof_coordinates"),
                                           dims=(len(points), gdim))

        # Input argument
        coordinate_dofs = L.Symbol("coordinate_dofs")

        # Loop indices
        i = L.Symbol("i")
        k = L.Symbol("k")
        ip = L.Symbol("ip")

        # Basis symbol
        phi = L.Symbol("phi")

        # TODO: Get rid of all places that use reference_to_physical_map, it is restricted to a basis of degree 1
        # Create code for evaluating coordinate mapping
        num_scalar_xdofs = _num_vertices(cell_shape)
        cg1_basis = reference_to_physical_map(cell_shape)
        phi_values = numpy.asarray([phi_comp for X in points for phi_comp in cg1_basis(X)])
        assert len(phi_values) == len(points) * num_scalar_xdofs

        # TODO: Use precision parameter here
        phi_values = clamp_table_small_numbers(phi_values)

        code = [
            L.Assign(
                dof_coordinates[ip][i],
                sum(phi_values[ip*num_scalar_xdofs + k] * coordinate_dofs[gdim*k + i]
                    for k in range(num_scalar_xdofs))
            )
            for ip in range(len(points))
            for i in range(gdim)
        ]

        # FIXME: This code assumes an affine coordinate field.
        #        To get around that limitation, make this function take another argument
        #            const ufc::coordinate_mapping * cm
        #        and generate code like this:
        """
        index_type X[tdim*num_dofs];
        tabulate_dof_coordinates(X);
        cm->compute_physical_coordinates(x, X, coordinate_dofs);
        """

        return code

    def tabulate_reference_dof_coordinates(self, L, ir, parameters):
        # TODO: Change signature to avoid copy? E.g.
        # virtual const std::vector<double> & tabulate_reference_dof_coordinates() const = 0;
        # See integral::enabled_coefficients for example

        # TODO: ensure points is a numpy array,
        #   get tdim from points.shape[1],
        #   place points in ir directly instead of the subdict
        ir = ir["tabulate_dof_coordinates"]

        # Raise error if tabulate_reference_dof_coordinates is ill-defined
        if not ir:
            msg = "tabulate_reference_dof_coordinates is not defined for this element"
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        # Extract coordinates and cell dimension
        tdim = ir["tdim"]
        points = ir["points"]

        # Output argument
        reference_dof_coordinates = L.Symbol("reference_dof_coordinates")

        # Reference coordinates
        dof_X = L.Symbol("dof_X")
        dof_X_values = [X[jj] for X in points for jj in range(tdim)]
        decl = L.ArrayDecl("static const double", dof_X,
                           (len(points) * tdim,), values=dof_X_values)
        copy = L.MemCopy(dof_X, reference_dof_coordinates, tdim*len(points))

        code = [decl, copy]
        return code

    def evaluate_reference_basis(self, L, ir, parameters):
        data = ir["evaluate_basis"]
        if isinstance(data, string_types):
            msg = "evaluate_reference_basis: %s" % data
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        return generate_evaluate_reference_basis(L, data, parameters)

    def evaluate_reference_basis_derivatives(self, L, ir, parameters):
        data = ir["evaluate_basis"]
        if isinstance(data, string_types):
            msg = "evaluate_reference_basis_derivatives: %s" % data
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        return generate_evaluate_reference_basis_derivatives(L, data, parameters)

    def transform_reference_basis_derivatives(self, L, ir, parameters):
        data = ir["evaluate_basis"]
        if isinstance(data, string_types):
            msg = "transform_reference_basis_derivatives: %s" % data
            return generate_error(L, msg, parameters["convert_exceptions_to_warnings"])

        # Get some known dimensions
        #element_cellname = data["cellname"]
        gdim = data["geometric_dimension"]
        tdim = data["topological_dimension"]
        max_degree = data["max_degree"]
        reference_value_size = data["reference_value_size"]
        physical_value_size = data["physical_value_size"]
        num_dofs = len(data["dofs_data"])

        max_g_d = gdim**max_degree
        max_t_d = tdim**max_degree

        # Output arguments
        values_symbol = L.Symbol("values")

        # Input arguments
        order = L.Symbol("order")
        num_points = L.Symbol("num_points")  # FIXME: Currently assuming 1 point?
        reference_values = L.Symbol("reference_values")
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        K = L.Symbol("K")

        # Internal variables
        transform = L.Symbol("transform")

        # Indices, I've tried to use these for a consistent purpose
        ip = L.Symbol("ip") # point
        i = L.Symbol("i")   # physical component
        j = L.Symbol("j")   # reference component
        k = L.Symbol("k")   # order
        r = L.Symbol("r")   # physical derivative number
        s = L.Symbol("s")   # reference derivative number
        d = L.Symbol("d")   # dof

        combinations_code = []
        if max_degree == 0:
            # Don't need combinations
            num_derivatives_t = 1  # TODO: I think this is the right thing to do to make this still work for order=0?
            num_derivatives_g = 1
        elif tdim == gdim:
            num_derivatives_t = L.Symbol("num_derivatives")
            num_derivatives_g = num_derivatives_t
            combinations_code += [
                L.VariableDecl("const " + index_type, num_derivatives_t,
                               L.Call("std::pow", (tdim, order))),
            ]

            # Add array declarations of combinations
            combinations_code_t, combinations_t = _generate_combinations(L, tdim, max_degree, order, num_derivatives_t)
            combinations_code += combinations_code_t
            combinations_g = combinations_t
        else:
            num_derivatives_t = L.Symbol("num_derivatives_t")
            num_derivatives_g = L.Symbol("num_derivatives_g")
            combinations_code += [
                L.VariableDecl("const " + index_type, num_derivatives_t,
                               L.Call("std::pow", (tdim, order))),
                L.VariableDecl("const " + index_type, num_derivatives_g,
                               L.Call("std::pow", (gdim, order))),
            ]
            # Add array declarations of combinations
            combinations_code_t, combinations_t = _generate_combinations(L, tdim, max_degree, order, num_derivatives_t, suffix="_t")
            combinations_code_g, combinations_g = _generate_combinations(L, gdim, max_degree, order, num_derivatives_g, suffix="_g")
            combinations_code += combinations_code_t
            combinations_code += combinations_code_g

        # Define expected dimensions of argument arrays
        J = L.FlattenedArray(J, dims=(num_points, gdim, tdim))
        detJ = L.FlattenedArray(detJ, dims=(num_points,))
        K = L.FlattenedArray(K, dims=(num_points, tdim, gdim))

        values = L.FlattenedArray(values_symbol,
            dims=(num_points, num_dofs, num_derivatives_g, physical_value_size))
        reference_values = L.FlattenedArray(reference_values,
            dims=(num_points, num_dofs, num_derivatives_t, reference_value_size))

        # Generate code to compute the derivative transform matrix
        transform_matrix_code = [
            # Initialize transform matrix to all 1.0
            L.ArrayDecl("double", transform, (max_g_d, max_t_d)),
            L.ForRanges(
                (r, 0, num_derivatives_g),
                (s, 0, num_derivatives_t),
                index_type=index_type,
                body=L.Assign(transform[r, s], 1.0)
            ),
            ]
        if max_degree > 0:
            transform_matrix_code += [
                # Compute transform matrix entries, each a product of K entries
                L.ForRanges(
                    (r, 0, num_derivatives_g),
                    (s, 0, num_derivatives_t),
                    (k, 0, order),
                    index_type=index_type,
                    body=L.AssignMul(transform[r, s],
                                     K[ip, combinations_t[s, k], combinations_g[r, k]])
                ),
            ]

        # Initialize values to 0, will be added to inside loops
        values_init_code = [
            L.MemZero(values_symbol, num_points * num_dofs * num_derivatives_g * physical_value_size),
            ]

        # Make offsets available in generated code
        reference_offsets = L.Symbol("reference_offsets")
        physical_offsets = L.Symbol("physical_offsets")
        dof_attributes_code = [
            L.ArrayDecl("const " + index_type, reference_offsets, (num_dofs,),
                        values=[dof_data["reference_offset"] for dof_data in data["dofs_data"]]),
            L.ArrayDecl("const " + index_type, physical_offsets, (num_dofs,),
                        values=[dof_data["physical_offset"] for dof_data in data["dofs_data"]]),
            ]

        # Build dof lists for each mapping type
        mapping_dofs = defaultdict(list)
        for idof, dof_data in enumerate(data["dofs_data"]):
            mapping_dofs[dof_data["mapping"]].append(idof)

        # Generate code for each mapping type
        d = L.Symbol("d")
        transform_apply_code = []
        for mapping in sorted(mapping_dofs):
            # Get list of dofs using this mapping
            idofs = mapping_dofs[mapping]

            # Select iteration approach over dofs
            if idofs == list(range(idofs[0], idofs[-1]+1)):
                # Contiguous
                dofrange = (d, idofs[0], idofs[-1]+1)
                idof = d
            else:
                # Stored const array of dof indices
                idofs_symbol = L.Symbol("%s_dofs" % mapping.replace(" ", "_"))
                dof_attributes_code += [
                    L.ArrayDecl("const " + index_type, idofs_symbol,
                                (len(idofs),), values=idofs),
                ]
                dofrange = (d, 0, len(idofs))
                idof = idofs_symbol[d]

            # NB! Array access to offsets, these are not Python integers
            reference_offset = reference_offsets[idof]
            physical_offset = physical_offsets[idof]

            # How many components does each basis function with this mapping have?
            # This should be uniform, i.e. there should be only one element in this set:
            num_reference_components, = set(data["dofs_data"][i]["num_components"] for i in idofs)

            M_scale, M_row, num_physical_components = generate_element_mapping(
                mapping, i,
                num_reference_components, tdim, gdim,
                J[ip], detJ[ip], K[ip]
            )

#            transform_apply_body = [
#                L.AssignAdd(values[ip, idof, r, physical_offset + k],
#                            transform[r, s] * reference_values[ip, idof, s, reference_offset + k])
#                for k in range(num_physical_components)
#            ]

            msg = "Using %s transform to map values back to the physical element." % mapping.replace("piola", "Piola")

            mapped_value = L.Symbol("mapped_value")
            transform_apply_code += [
                L.ForRanges(
                    dofrange,
                    (s, 0, num_derivatives_t),
                    (i, 0, num_physical_components),
                    index_type=index_type, body=[
                        # Unrolled application of mapping to one physical component,
                        # for affine this automatically reduces to
                        #   mapped_value = reference_values[..., reference_offset]
                        L.Comment(msg),
                        L.VariableDecl("const double", mapped_value,
                                       M_scale * sum(M_row[jj] * reference_values[ip, idof, s, reference_offset + jj]
                                                     for jj in range(num_reference_components))),
                        # Apply derivative transformation, for order=0 this reduces to
                        # values[ip,idof,0,physical_offset+i] = transform[0,0]*mapped_value
                        L.Comment("Mapping derivatives back to the physical element"),
                        L.ForRanges(
                            (r, 0, num_derivatives_g),
                            index_type=index_type, body=[
                                L.AssignAdd(values[ip, idof, r, physical_offset + i],
                                            transform[r, s] * mapped_value)
                        ])
                ])
            ]

        # Transform for each point
        point_loop_code = [
            L.ForRange(ip, 0, num_points, index_type=index_type, body=(
                transform_matrix_code
                + transform_apply_code
            ))
        ]

        # Join code
        code = (
            combinations_code
            + values_init_code
            + dof_attributes_code
            + point_loop_code
        )
        return code


def _num_vertices(cell_shape):
    """Returns number of vertices for a given cell shape."""

    num_vertices_dict = {"interval": 2, "triangle": 3, "tetrahedron": 4, "quadrilateral": 4, "hexahedron": 8}
    return num_vertices_dict[cell_shape]





# CURRENTLY UNUSED CODE, CONVERT TO CREATE NEW ufc::finite_element::apply_mappings functions


def _generate_apply_mapping_to_computed_values(L, dof_data):
    mapping = dof_data["mapping"]
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]

    #physical_values[num_points][num_dofs][physical_value_size]
    #reference_values[num_points][num_dofs][reference_value_size]

    code = []

    # FIXME: Define values numbering
    if mapping == "affine":
        # Just copy values
        if num_components == 1:
            code += [
                L.ForRange(ip, 0, num_points, body=
                    L.Assign(physical_values[ip][physical_offset],
                             reference_values[ip][reference_offset])),
                ]
        else:
            code += [
                L.ForRange(ip, 0, num_points, body=
                    L.ForRange(k, 0, num_components, body=
                        L.Assign(physical_values[ip][physical_offset + k],
                                 reference_values[ip][reference_offset + k]))),
                ]
        return code

    else:
        FIXME

    return code


def __ffc_implementation_of__generate_apply_mapping_to_computed_values(L):
    # Apply transformation if applicable.
    mapping = dof_data["mapping"]
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]

    if mapping == "affine":
        pass

    elif mapping == "contravariant piola":
        code += ["", f_comment("Using contravariant Piola transform to map values back to the physical element")]

        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]

        # Create names for inner product.
        basis_col = [f_tmp_ref(j) for j in range(tdim)]
        for i in range(gdim):
            # Create Jacobian.
            jacobian_row = [f_trans("J", i, j, gdim, tdim, None) for j in range(tdim)]
            # Create inner product and multiply by inverse of Jacobian.
            inner = f_group(f_inner(jacobian_row, basis_col)) # sum_j J[i,j], values[offset + j])
            value = f_mul([f_inv(f_detJ(None)), inner])
            name = f_component(f_values, i + offset)
            # FIXME: This is writing values[offset+:] = M[:,:] * values[offset+:],
            #        i.e. offset must be physical (unless there's a bug).
            #        We want to use reference offset for evaluate_reference_basis,
            #        and to make this mapping read from one array using reference offset
            #        and write to another array using physical offset.
            code += [f_assign(name, value)]

    elif mapping == "covariant piola":
        code += ["", f_comment("Using covariant Piola transform to map values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(tdim)]
        for i in range(gdim):
            # Create inverse of Jacobian.
            inv_jacobian_column = [f_trans("JINV", j, i, tdim, gdim, None) for j in range(tdim)]

            # Create inner product of basis values and inverse of Jacobian.
            value = f_group(f_inner(inv_jacobian_column, basis_col))
            name = f_component(f_values, i + offset)
            code += [f_assign(name, value)]

    elif mapping == "double covariant piola":
        code += ["", f_comment("Using double covariant Piola transform to map values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim  # noqa: E741
            # g_il = K_ji G_jk K_kl
            value = f_group(f_inner(
                [f_inner([f_trans("JINV", j, i, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_trans("JINV", k, l, tdim, gdim, None)
                 for k in range(tdim)]))
            name = f_component(f_values, p + offset)
            code += [f_assign(name, value)]

    elif mapping == "double contravariant piola":
        code += ["", f_comment("Pullback of a matrix-valued funciton as contravariant 2-tensor mapping values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim  # noqa: E741
            # g_il = (detJ)^(-2) J_ij G_jk J_lk
            value = f_group(f_inner(
                [f_inner([f_trans("J", i, j, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_trans("J", l, k, tdim, gdim, None)
                 for k in range(tdim)]))
            value = f_mul([f_inv(f_detJ(None)), f_inv(f_detJ(None)), value])
            name = f_component(f_values, p + offset)
            code += [f_assign(name, value)]

    else:
        error("Unknown mapping: %s" % mapping)



# -*- coding: utf-8 -*-
"""Work in progress translation of FFC evaluatebasisderivatives code to uflacs CNodes format."""

from six import string_types
import numpy
import math

from ffc.log import error
from ffc.uflacs.backends.ufc.utils import generate_error
from ffc.uflacs.backends.ufc.evaluatebasis import _generate_compute_basisvalues, tabulate_coefficients
from ffc.uflacs.backends.ufc.evalderivs import _generate_combinations
from ffc.uflacs.backends.ufc.jacobian import jacobian, inverse_jacobian, orientation, fiat_coordinate_mapping, _mapping_transform

# Used for various indices and arrays in this file
index_type = "std::size_t"

def _compute_reference_derivatives(L, data, dof_data):
    """Compute derivatives on the reference element by recursively
    multiply coefficients with the relevant derivatives of the
    polynomial base until the requested order of derivatives has been
    reached. After this take the dot product with the basisvalues.

    """

    # Prefetch formats to speed up code generation

    tdim = data["topological_dimension"]
    gdim = data["geometric_dimension"]
    max_degree = data["max_degree"]

    _t, _g = ("","") if (tdim == gdim) else ("_t", "_g")
    num_derivs_t = L.Symbol("num_derivatives" + _t)
    num_derivs_g = L.Symbol("num_derivatives" + _g)

    # Get number of components.
    num_components = dof_data["num_components"]

    # Get shape of derivative matrix (they should all have the same
    # shape) and verify that it is a square matrix.
    shape_dmats = dof_data["dmats"][0].shape
    if shape_dmats[0] != shape_dmats[1]:
        error("Something is wrong with the dmats:\n%s" % str(dof_data["dmats"]))

    code = [L.Comment("Compute reference derivatives.")]

    # Declare pointer to array that holds derivatives on the FIAT element
    code += [L.Comment("Declare array of derivatives on FIAT element.")]
    # The size of the array of reference derivatives is equal to the
    # number of derivatives times the number of components of the
    # basis element
    num_vals = num_components*num_derivs_t
    nds = tdim**max_degree * num_components

    mapping = dof_data["mapping"]
    if "piola" in mapping and "double" not in mapping and gdim > num_components :
        # In either of the Piola cases, the value space of the
        # derivatives is the geometric dimension rather than the
        # topological dimension.  Increase size of derivatives array
        # if needed.
        nds = tdim**max_degree * gdim

    derivatives = L.Symbol("derivatives")
    code += [L.ArrayDecl("double", derivatives, nds, 0.0)]

    # Declare matrix of dmats (which will hold the matrix product of
    # all combinations) and dmats_old which is needed in order to
    # perform the matrix product.
    code += [L.Comment("Declare derivative matrix (of polynomial basis).")]
    value = numpy.eye(shape_dmats[0])
    dmats = L.Symbol("dmats")
    code += [L.ArrayDecl("double", dmats, shape_dmats, value)]
    code += [L.Comment("Declare (auxiliary) derivative matrix (of polynomial basis).")]
    dmats_old = L.Symbol("dmats_old")
    code += [L.ArrayDecl("double", dmats_old, shape_dmats, value)]

    r = L.Symbol("r")
    s = L.Symbol("s")
    t = L.Symbol("t")
    n = L.Symbol("n")

    lines = [L.Comment("Reset dmats to identity"),
             L.MemZero(L.AddressOf(dmats[0][0]), shape_dmats[0]*shape_dmats[1]),
             L.ForRange(t, 0, shape_dmats[0], index_type=index_type, body=[L.Assign(dmats[t][t], 1.0)])]

    lines_s = [L.MemCopy(L.AddressOf(dmats[0][0]), L.AddressOf(dmats_old[0][0]), shape_dmats[0]*shape_dmats[1]),
               L.MemZero(L.AddressOf(dmats[0][0]), shape_dmats[0]*shape_dmats[1])]

    lines_s += [L.Comment("Update dmats using an inner product.")]

    # Create dmats matrix by multiplication
    comb = L.Symbol("combinations" + _t)

    for i in range(len(dof_data["dmats"])):
        lines_s += [L.Comment("_dmats_product(shape_dmats, comb[r][s], %d)" % i)]
        dmats_i = L.Symbol("dmats%d" % i)
        u = L.Symbol("u")
        tu = L.Symbol("tu")
        lines_comb = [L.ForRange(t, 0, shape_dmats[0], index_type=index_type,
              body = [L.ForRange(u, 0, shape_dmats[1], index_type=index_type,
              body = [L.ForRange(tu, 0, shape_dmats[1], index_type=index_type,
              body = [L.AssignAdd(dmats[t][u], dmats_old[tu][u] * dmats_i[t][tu])])])])]

        lines_s += [L.If(L.EQ(comb[n - 1][r][s], i), lines_comb)]

    lines += [L.Comment("Looping derivative order to generate dmats."),
              L.ForRange(s, 0, n, index_type=index_type, body=lines_s)]

    # Compute derivatives for all components
    lines_c = []
    basisvalues = L.Symbol("basisvalues")
    for i in range(num_components):
        coeffs = L.Symbol("coefficients%d" % i)
        lines_c += [L.AssignAdd(derivatives[i*num_derivs_t + r], coeffs[s]*dmats[s,t]*basisvalues[t])]
    lines += [L.ForRange(s, 0, shape_dmats[0], index_type=index_type,
        body=[L.ForRange(t, 0, shape_dmats[1], index_type=index_type, body=lines_c)])]

    lines += _mapping_transform(L, data, dof_data, derivatives, r, num_derivs_t)

    code += [L.Comment("Loop possible derivatives."),
             L.ForRange(r, 0, num_derivs_t, index_type=index_type, body=lines)]

    return code


def _transform_derivatives(L, data, dof_data):
    """Transform derivatives back to the physical element by applying the
    transformation matrix."""

    tdim = data["topological_dimension"]
    gdim = data["geometric_dimension"]

    _t, _g = ("","") if (tdim == gdim) else ("_t", "_g")
    num_derivs_t = L.Symbol("num_derivatives" + _t)
    num_derivs_g = L.Symbol("num_derivatives" + _g)

    # Get number of components and offset.
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]
    offset = reference_offset  # physical_offset # FIXME: Should be physical offset but that breaks tests

    mapping = dof_data["mapping"]
    if "piola" in mapping and "double" not in mapping:
        # In either of the Piola cases, the value space of the derivatives
        # is the geometric dimension rather than the topological dimension.
        num_components = gdim

    code = [L.Comment("Transform derivatives back to physical element")]

    lines = []
    r = L.Symbol("r")
    s = L.Symbol("s")
    values = L.Symbol("values")
    transform = L.Symbol("transform")
    derivatives = L.Symbol("derivatives")
    for i in range(num_components):
        lines += [L.AssignAdd(values[(offset + i)*num_derivs_g + r],
                              transform[r, s]*derivatives[i*num_derivs_t + s])]

    code += [L.ForRange(r, 0, num_derivs_g, index_type=index_type,
                        body=[L.ForRange(s, 0, num_derivs_t, index_type=index_type, body=lines)])]

    return code


def _tabulate_dmats(L, dof_data):
    "Tabulate the derivatives of the polynomial base"

    # Get derivative matrices (coefficients) of basis functions,
    # computed by FIAT at compile time.

    code = [L.Comment("Tables of derivatives of the polynomial base (transpose).")]

    # Generate tables for each spatial direction.
    for i, dmat in enumerate(dof_data["dmats"]):

        # Extract derivatives for current direction (take transpose,
        # FIAT_NEW PolynomialSet.tabulate()).
        matrix = numpy.transpose(dmat)

        # Get shape and check dimension (This is probably not needed).
        #        shape = numpy.shape(matrix)
        if not (matrix.shape[0] == matrix.shape[1] == dof_data["num_expansion_members"]):
            error("Something is wrong with the shape of dmats.")

        # Declare varable name for coefficients.
        table = L.Symbol("dmats%d" % i)
        code += [L.ArrayDecl("static const double", table, matrix.shape, matrix)]

    return code


def _generate_dof_code(L, data, dof_data):
    "Generate code for a basis."

    basisvalues = L.Symbol("basisvalues")
    Y = L.Symbol("Y")
    element_cellname = data["cellname"]
    embedded_degree = dof_data["embedded_degree"]
    num_members = dof_data["num_expansion_members"]
    code = _generate_compute_basisvalues(L, basisvalues, Y, element_cellname, embedded_degree, num_members)

    # Tabulate coefficients.
    code += tabulate_coefficients(L, dof_data)

    # Tabulate coefficients for derivatives.
    code += _tabulate_dmats(L, dof_data)

    # Compute the derivatives of the basisfunctions on the reference
    # (FIAT) element, as the dot product of the new coefficients and
    # basisvalues.
    code += _compute_reference_derivatives(L, data, dof_data)

    # Transform derivatives to physical element by multiplication with
    # the transformation matrix.
    code += _transform_derivatives(L, data, dof_data)

    return code


def _generate_transform(L, element_cellname, gdim, tdim, max_degree):
    """Generate the transformation matrix, which is used to transform
    derivatives from reference element back to the physical
    element.

    """

    max_g_d = gdim**max_degree
    max_t_d = tdim**max_degree

    K = L.Symbol("K")
    transform = L.Symbol("transform")
    col = L.Symbol("col")
    row = L.Symbol("row")

    _t, _g = ("","") if (tdim == gdim) else ("_t", "_g")
    comb_t = L.Symbol("combinations" + _t)
    comb_g = L.Symbol("combinations" + _g)
    num_derivatives_t = L.Symbol("num_derivatives" + _t)
    num_derivatives_g = L.Symbol("num_derivatives" + _g)

    K = L.FlattenedArray(K, dims=(tdim, gdim))

    code = [L.Comment("Declare transformation matrix")]
    code += [L.ArrayDecl("double", transform, (max_g_d, max_t_d), numpy.ones((max_g_d, max_t_d)))]
    code += [L.Comment("Construct transformation matrix")]
    k = L.Symbol("k")
    n = L.Symbol("n")
    inner_loop = L.ForRange(col, 0, num_derivatives_t, index_type=index_type,
           body=[L.ForRange(k, 0, n, index_type=index_type,
           body=[L.AssignMul(transform[row][col], K[comb_t[n-1][col][k], comb_g[n-1][row][k]])])])

    code += [L.ForRange(row, 0, num_derivatives_g, index_type=index_type, body=inner_loop)]

    return code


def generate_evaluate_basis_derivatives(L, data):
    """Evaluate the derivatives of an element basisfunction at a
    point. The values are computed as in FIAT as the matrix product of
    the coefficients (computed at compile time), basisvalues which are
    dependent on the coordinate and thus have to be computed at run
    time and combinations (depending on the order of derivative) of
    dmats tables which hold the derivatives of the expansion
    coefficients.

    """

    if isinstance(data, string_types):
        msg = "evaluate_basis_derivatives: %s" % data
        return [generate_error(L, msg, True)]

    # Initialise return code.
    code = []

    # Get the element cell domain, geometric and topological dimension.
    element_cellname = data["cellname"]
    gdim = data["geometric_dimension"]
    tdim = data["topological_dimension"]
    max_degree = data["max_degree"]
    physical_value_size = data["physical_value_size"]

    # Compute number of derivatives that has to be computed, and
    # declare an array to hold the values of the derivatives on the
    # reference element.
    values = L.Symbol("values")
    n = L.Symbol("n")
    dofs = L.Symbol("dofs")
    x = L.Symbol("x")
    coordinate_dofs = L.Symbol("coordinate_dofs")
    cell_orientation = L.Symbol("cell_orientation")

    if tdim == gdim:
        _t, _g = ("", "")
        num_derivatives_t = L.Symbol("num_derivatives")
        num_derivatives_g = L.Symbol("num_derivatives")
        code += [L.VariableDecl("std::size_t", num_derivatives_t, L.Call("std::pow", (tdim, n)))]
    else:
        _t, _g = ("_t", "_g")
        num_derivatives_t = L.Symbol("num_derivatives_t")
        num_derivatives_g = L.Symbol("num_derivatives_g")
        if max_degree > 0:
            code += [L.VariableDecl("std::size_t", num_derivatives_t, L.Call("std::pow", (tdim, n)))]
        code += [L.VariableDecl("std::size_t", num_derivatives_g, L.Call("std::pow", (gdim, n)))]

    # Reset all values.
    code += [L.MemZero(values, physical_value_size*num_derivatives_g)]

    # Handle values of argument 'n'.
    code += [L.Comment("Call evaluate_basis_all if order of derivatives is equal to zero.")]
    code += [L.If(L.EQ(n, 0), [L.Call("evaluate_basis", (L.Symbol("i"), values, x, coordinate_dofs, cell_orientation)), L.Return()])]

    # If max_degree is zero, return code (to avoid declarations such as
    # combinations[1][0]) and because there's nothing to compute.)
    if max_degree == 0:
        return code

    code += [L.Comment("If order of derivatives is greater than the maximum polynomial degree, return zeros.")]
    code += [L.If(L.GT(n, max_degree), [L.Return()])]

    # Generate geo code.
    code += jacobian(L, gdim, tdim, element_cellname)
    code += inverse_jacobian(L, gdim, tdim, element_cellname)
    if data["needs_oriented"] and tdim != gdim:
        code += orientation(L)

    code += fiat_coordinate_mapping(L, element_cellname, gdim)

    # Generate all possible combinations of derivatives.
    combinations_code_t, combinations_t = _generate_combinations(L, tdim, max_degree, n, num_derivatives_t, _t)
    code += combinations_code_t
    if tdim != gdim:
        combinations_code_g, combinations_g = _generate_combinations(L, gdim, max_degree, n, num_derivatives_g, "_g")
        code += combinations_code_g

    # Generate the transformation matrix.
    code += _generate_transform(L, element_cellname, gdim, tdim, max_degree)

    # Create code for all basis values (dofs).
    dof_cases = []
    for i, dof_data in enumerate(data["dofs_data"]):
        dof_cases.append((i, _generate_dof_code(L, data, dof_data)))
    code += [L.Switch(L.Symbol("i"), dof_cases)]
    return code

def generate_evaluate_basis_derivatives_all(L, data):
    """Like evaluate_basis, but return the values of all basis
    functions (dofs)."""

    if isinstance(data, string_types):
        msg = "evaluate_basis_derivatives_all: %s" % data
        return [generate_error(L, msg, True)]

    # Initialise return code
    code = []

    # FIXME: KBO: Figure out which return format to use, either:
    # [dN0[0]/dx, dN0[0]/dy, dN0[1]/dx, dN0[1]/dy, dN1[0]/dx,
    # dN1[0]/dy, dN1[1]/dx, dN1[1]/dy, ...]
    # or
    # [dN0[0]/dx, dN1[0]/dx, ..., dN0[1]/dx, dN1[1]/dx, ...,
    # dN0[0]/dy, dN1[0]/dy, ..., dN0[1]/dy, dN1[1]/dy, ...]
    # or
    # [dN0[0]/dx, dN0[1]/dx, ..., dN1[0]/dx, dN1[1]/dx, ...,
    # dN0[0]/dy, dN0[1]/dy, ..., dN1[0]/dy, dN1[1]/dy, ...]
    # for vector (tensor elements), currently returning option 1.

    # FIXME: KBO: For now, just call evaluate_basis_derivatives and
    # map values accordingly, this will keep the amount of code at a
    # minimum. If it turns out that speed is an issue (overhead from
    # calling evaluate_basis), we can easily generate all the code.

    # Get total value shape and space dimension for entire element
    # (possibly mixed).
    physical_value_size = data["physical_value_size"]
    space_dimension = data["space_dimension"]
    max_degree = data["max_degree"]
    gdim = data["geometric_dimension"]
    tdim = data["topological_dimension"]

    n = L.Symbol("n")
    x = L.Symbol("x")
    coordinate_dofs = L.Symbol("coordinate_dofs")
    cell_orientation = L.Symbol("cell_orientation")
    values = L.Symbol("values")

    # Special case where space dimension is one (constant elements).
    if space_dimension == 1:
        code += [L.Comment("Element is constant, calling evaluate_basis_derivatives.")]
        code += [L.Call("evaluate_basis_derivatives", (0, n, values, x, coordinate_dofs, cell_orientation))]
        return code

    # Compute number of derivatives.
    if tdim == gdim:
        num_derivatives = L.Symbol("num_derivatives")
    else:
        num_derivatives = L.Symbol("num_derivatives_g")

    # If n == 0, call evaluate_basis.
    code += [L.Comment("Call evaluate_basis_all if order of derivatives is equal to zero.")]
    code += [L.If(L.EQ(n, 0), [L.Call("evaluate_basis_all", (values, x, coordinate_dofs, cell_orientation)), L.Return()])]

    code += [L.VariableDecl("unsigned int", num_derivatives, L.Call("std::pow", (gdim, n)))]

    num_vals = physical_value_size * num_derivatives

    # Reset values.
    code += [L.Comment("Set values equal to zero.")]
    code += [L.MemZero(values, num_vals*space_dimension)]

    # If n > max_degree, return zeros.
    code += [L.Comment("If order of derivatives is greater than the maximum polynomial degree, return zeros.")]
    code += [L.If(L.GT(n, max_degree), [L.Return()])]

    # Declare helper value to hold single dof values and reset.
    code += [L.Comment("Helper variable to hold values of a single dof.")]
    nds = gdim**max_degree * physical_value_size
    dof_values = L.Symbol("dof_values")
    code += [L.ArrayDecl("double", dof_values, (nds,), 0.0)]

    # Create loop over dofs that calls evaluate_basis_derivatives for
    # a single dof and inserts the values into the global array.
    code += [L.Comment("Loop dofs and call evaluate_basis_derivatives.")]

    values = L.FlattenedArray(values, dims=(space_dimension, num_vals))
    r = L.Symbol("r")
    s = L.Symbol("s")
    loop_s = L.ForRange(s, 0, num_vals, index_type=index_type, body=[L.Assign(values[r, s], dof_values[s])])

    code += [L.ForRange(r, 0, space_dimension, index_type=index_type, body=[L.Call("evaluate_basis_derivatives", (r, n, dof_values, x, coordinate_dofs, cell_orientation)),
                                                                            loop_s])]
    return code



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2017 Anders Logg and Martin Sandve Alnæs, Chris Richardson
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.


# Note: Much of the code in this file is a direct translation
# from the old implementation in FFC, although some improvements
# have been made to the generated code.

from ffc.uflacs.backends.ufc.utils import generate_error
from ffc.uflacs.backends.ufc.jacobian import jacobian, inverse_jacobian, orientation
from ufl.permutation import build_component_numbering
from ffc.utils import pick_first

index_type="std::size_t"

def reference_to_physical_map(cellname):
    "Returns a map from reference coordinates to physical element coordinates"

    if cellname == "interval":
        return lambda x: (1.0 - x[0], x[0])
    elif cellname == "triangle":
        return lambda x: (1.0 - x[0] - x[1], x[0], x[1])
    elif cellname == "tetrahedron":
        return lambda x: (1.0 - x[0] - x[1] - x[2], x[0], x[1], x[2])
    elif cellname == "quadrilateral":
        return lambda x: ((1-x[0])*(1-x[1]), (1-x[0])*x[1], x[0]*(1-x[1]), x[0]*x[1])
    elif cellname == "hexahedron":
        return lambda x: ((1-x[0])*(1-x[1])*(1-x[2]),
            (1-x[0])*(1-x[1])*x[2],
            (1-x[0])*x[1]*(1-x[2]),
            (1-x[0])*x[1]*x[2],
            x[0]*(1-x[1])*(1-x[2]),
            x[0]*(1-x[1])*x[2],
            x[0]*x[1]*(1-x[2]),
            x[0]*x[1]*x[2])


def _change_variables(L, mapping, gdim, tdim, offset):
    """Generate code for mapping function values according to
    'mapping' and offset.

    The basics of how to map a field from a physical to the reference
    domain. (For the inverse approach -- see interpolatevertexvalues)

    Let g be a field defined on a physical domain T with physical
    coordinates x. Let T_0 be a reference domain with coordinates
    X. Assume that F: T_0 -> T such that

      x = F(X)

    Let J be the Jacobian of F, i.e J = dx/dX and let K denote the
    inverse of the Jacobian K = J^{-1}. Then we (currently) have the
    following four types of mappings:

    'affine' mapping for g:

      G(X) = g(x)

    For vector fields g:

    'contravariant piola' mapping for g:

      G(X) = det(J) K g(x)   i.e  G_i(X) = det(J) K_ij g_j(x)

    'covariant piola' mapping for g:

      G(X) = J^T g(x)          i.e  G_i(X) = J^T_ij g(x) = J_ji g_j(x)

    'double covariant piola' mapping for g:

      G(X) = J^T g(x) J     i.e. G_il(X) = J_ji g_jk(x) J_kl

    'double contravariant piola' mapping for g:

      G(X) = det(J)^2 K g(x) K^T  i.e. G_il(X)=(detJ)^2 K_ij g_jk K_lk

    """

    # meg: Various mappings must be handled both here and in
    # interpolate_vertex_values. Could this be abstracted out?

    values = L.Symbol("vals")

    if mapping == "affine":
        return [values[offset]]
    elif mapping == "contravariant piola":
        # Map each component from physical to reference using inverse
        # contravariant piola
        detJ = L.Symbol("detJ")
        K = L.Symbol("K")
        K = L.FlattenedArray(K, dims=(tdim, gdim))
        w = []
        for i in range(tdim):
            inner = 0.0
            for j in range(gdim):
                inner += values[j + offset]*K[i, j]
            w.append(inner*detJ)
        return w

    elif mapping == "covariant piola":
        # Map each component from physical to reference using inverse
        # covariant piola
        detJ = L.Symbol("detJ")
        J = L.Symbol("J")
        J = L.FlattenedArray(J, dims=(gdim, tdim))
        w = []
        for i in range(tdim):
            inner = 0.0
            for j in range(gdim):
                inner += values[j + offset]*J[j, i]
            w.append(inner)
        return w

    elif mapping == "double covariant piola":
        # physical to reference pullback as a covariant 2-tensor
        w = []
        J = L.Symbol("J")
        J = L.FlattenedArray(J, dims=(gdim, tdim))
        for i in range(tdim):
            for l in range(tdim):
                inner = 0.0
                for k in range(gdim):
                    for j in range(gdim):
                        inner += J[j, i] * values[j*tdim + k + offset] * J[k, l]
                w.append(inner)
        return w

    elif mapping == "double contravariant piola":
        # physical to reference using double contravariant piola
        w = []
        K = L.Symbol("K")
        K = L.FlattenedArray(K, dims=(tdim, gdim))
        detJ = L.Symbol("detJ")
        for i in range(tdim):
            for l in range(tdim):
                inner = 0.0
                for k in range(gdim):
                    for j in range(gdim):
                        inner += K[i, j] * values[j*tdim + k + offset] * K[l, k]
                w.append(inner*detJ*detJ)
        return w

    else:
        raise Exception("The mapping (%s) is not allowed" % mapping)

def _generate_body(L, i, dof, mapping, gdim, tdim, cell_shape, offset=0):
    "Generate code for a single dof."

    # EnrichedElement is handled by having [None, ..., None] dual basis
    if not dof:
        msg = "evaluate_dof(s) for enriched element not implemented."
        return ([generate_error(L, msg, True)], 0.0)

    points = list(dof.keys())

    # Generate different code if multiple points. (Otherwise ffc
    # compile time blows up.)
    if len(points) > 1:
        return _generate_multiple_points_body(L, i, dof, mapping, gdim, tdim,
                                              offset)

    # Get weights for mapping reference point to physical
    x = points[0]
    w = reference_to_physical_map(cell_shape)(x)

    # Map point onto physical element: y = F_K(x)
    code = []

    y = L.Symbol("y")
    coordinate_dofs = L.Symbol("coordinate_dofs")
    vals = L.Symbol("vals")
    c = L.Symbol("c")
    for j in range(gdim):
        yy = 0.0
        for k in range(len(w)):
            yy += w[k]*coordinate_dofs[k*gdim + j]
        code += [L.Assign(y[j], yy)]

    # Evaluate function at physical point
    code += [L.Call("f.evaluate", (vals, y, c))]

    # Map function values to the reference element
    F = _change_variables(L, mapping, gdim, tdim, offset)

    # Simple affine functions deserve special case:
    if len(F) == 1:
        return (code, dof[x][0][0]*F[0])

    # Flatten multi-indices
    (index_map, _) = build_component_numbering([tdim] * len(dof[x][0][1]), ())
    # Take inner product between components and weights
    value = 0.0
    for (w, k) in dof[x]:
        value += w*F[index_map[k]]

    # Return eval code and value
    return (code, value)

def _generate_multiple_points_body(L, i, dof, mapping, gdim, tdim,
                                   offset=0):
    "Generate c++ for-loop for multiple points (integral bodies)"

    result = L.Symbol("result")
    code = [L.Assign(result, 0.0)]
    points = list(dof.keys())
    n = len(points)

    # Get number of tokens per point
    tokens = [dof[x] for x in points]
    len_tokens = pick_first([len(t) for t in tokens])

    # Declare points
    #    points = format["list"]([format["list"](x) for x in points])
    X_i = L.Symbol("X_%d" % i)
    code += [L.ArrayDecl("double", X_i, [n, tdim], points)]

    # Declare components
    components = [[c[0] for (w, c) in token] for token in tokens]
    #    components = format["list"]([format["list"](c) for c in components])
    D_i = L.Symbol("D_%d" % i)
    code += [L.ArrayDecl("int", D_i, [n, len_tokens], components)]

    # Declare weights
    weights = [[w for (w, c) in token] for token in tokens]
    #    weights = format["list"]([format["list"](w) for w in weights])
    W_i = L.Symbol("W_%d" % i)
    code += [L.ArrayDecl("double", W_i, [n, len_tokens], weights)]

    # Declare copy variable:
    copy_i = L.Symbol("copy_%d" % i)
    code += [L.ArrayDecl("double", copy_i, tdim)]

    # Add loop over points
    code += [L.Comment("Loop over points")]

    # Map the points from the reference onto the physical element
    r = L.Symbol("r")
    w0 = L.Symbol("w0")
    w1 = L.Symbol("w1")
    w2 = L.Symbol("w2")
    w3 = L.Symbol("w3")
    y = L.Symbol("y")
    coordinate_dofs = L.Symbol("coordinate_dofs")

    if tdim == 1:
        lines_r = [L.Comment("Evaluate basis functions for affine mapping"),
                   L.VariableDecl("const double", w0, 1 - X_i[r][0]),
                   L.VariableDecl("const double", w1, X_i[r][0]),
                   L.Comment("Compute affine mapping y = F(X)")]
        for j in range(gdim):
            lines_r += [L.Assign(y[j],
                                 w0*coordinate_dofs[j] + w1*coordinate_dofs[j + gdim])]
    elif tdim == 2:
        lines_r = [L.Comment("Evaluate basis functions for affine mapping"),
                   L.VariableDecl("const double", w0, 1 - X_i[r][0] - X_i[r][1]),
                   L.VariableDecl("const double", w1, X_i[r][0]),
                   L.VariableDecl("const double", w2, X_i[r][1]),
                   L.Comment("Compute affine mapping y = F(X)")]
        for j in range(gdim):
            lines_r += [L.Assign(y[j],
                                 w0*coordinate_dofs[j]
                                 + w1*coordinate_dofs[j + gdim]
                                 + w2*coordinate_dofs[j + 2*gdim])]
    elif tdim == 3:
        lines_r = [L.Comment("Evaluate basis functions for affine mapping"),
                   L.VariableDecl("const double", w0, 1 - X_i[r][0] - X_i[r][1] - X_i[r][2]),
                   L.VariableDecl("const double", w1, X_i[r][0]),
                   L.VariableDecl("const double", w2, X_i[r][1]),
                   L.VariableDecl("const double", w3, X_i[r][2]),
                   L.Comment("Compute affine mapping y = F(X)")]
        for j in range(gdim):
            lines_r += [L.Assign(y[j],
                                 w0*coordinate_dofs[j]
                                 + w1*coordinate_dofs[j + gdim]
                                 + w2*coordinate_dofs[j + 2*gdim]
                                 + w3*coordinate_dofs[j + 3*gdim])]

    # Evaluate function at physical point
    lines_r += [L.Comment("Evaluate function at physical point")]
    y = L.Symbol("y")
    vals = L.Symbol("vals")
    c = L.Symbol("c")
    lines_r += [L.Call("f.evaluate",(vals, y, c))]

    # Map function values to the reference element
    lines_r += [L.Comment("Map function to reference element")]
    F = _change_variables(L, mapping, gdim, tdim, offset)
    lines_r += [L.Assign(copy_i[k], F_k)
                for (k, F_k) in enumerate(F)]

    # Add loop over directional components
    lines_r += [L.Comment("Loop over directions")]

    s = L.Symbol("s")
    lines_r += [L.ForRange(s, 0, len_tokens, index_type=index_type,
                       body=[L.AssignAdd(result, copy_i[D_i[r, s]] * W_i[r, s])])]

    # Generate loop over r and add to code.
    code += [L.ForRange(r, 0, n, index_type=index_type, body=lines_r)]
    return (code, result)

def generate_evaluate_dof(L, ir):
    "Generate code for evaluate_dof."

    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]

    cell_shape = ir["cell_shape"]

    # Enriched element, no dofs defined
    if not any(ir["dofs"]):
        code = []
    else:
        # Declare variable for storing the result and physical coordinates
        code = [L.Comment("Declare variables for result of evaluation")]
        vals = L.Symbol("vals")
        code += [L.ArrayDecl("double", vals, ir["physical_value_size"])]
        code += [L.Comment("Declare variable for physical coordinates")]
        y = L.Symbol("y")
        code += [L.ArrayDecl("double", y, gdim)]

        # Check whether Jacobians are necessary.
        needs_inverse_jacobian = any(["contravariant piola" in m
                                      for m in ir["mappings"]])
        needs_jacobian = any(["covariant piola" in m for m in ir["mappings"]])

        # Intermediate variable needed for multiple point dofs
        needs_temporary = any(dof is not None and len(dof) > 1 for dof in ir["dofs"])
        if needs_temporary:
            result = L.Symbol("result")
            code += [L.VariableDecl("double", result)]

        if needs_jacobian or needs_inverse_jacobian:
            code += jacobian(L, gdim, tdim, cell_shape)

        if needs_inverse_jacobian:
            code += inverse_jacobian(L, gdim, tdim, cell_shape)
            if tdim != gdim :
                code += orientation(L)

    # Extract variables
    mappings = ir["mappings"]
    offsets = ir["physical_offsets"]

    # Generate bodies for each degree of freedom
    cases = []
    for (i, dof) in enumerate(ir["dofs"]):
        c, r = _generate_body(L, i, dof, mappings[i], gdim, tdim, cell_shape, offsets[i])
        c += [L.Return(r)]
        cases.append((i, c))

    code += [L.Switch(L.Symbol("i"), cases)]
    code += [L.Return(0.0)]
    return code

def generate_evaluate_dofs(L, ir):
    "Generate code for evaluate_dofs."
    # FIXME: consolidate with evaluate_dofs
    # FIXME: replace

    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]

    cell_shape = ir["cell_shape"]

    # Enriched element, no dofs defined
    if not any(ir["dofs"]):
        code = []
    else:
        # Declare variable for storing the result and physical coordinates
        code = [L.Comment("Declare variables for result of evaluation")]
        vals = L.Symbol("vals")
        code += [L.ArrayDecl("double", vals, ir["physical_value_size"])]
        code += [L.Comment("Declare variable for physical coordinates")]
        y = L.Symbol("y")
        code += [L.ArrayDecl("double", y, gdim)]

        # Check whether Jacobians are necessary.
        needs_inverse_jacobian = any(["contravariant piola" in m
                                      for m in ir["mappings"]])
        needs_jacobian = any(["covariant piola" in m for m in ir["mappings"]])

        # Intermediate variable needed for multiple point dofs
        needs_temporary = any(dof is not None and len(dof) > 1 for dof in ir["dofs"])
        if needs_temporary:
            result = L.Symbol("result")
            code += [L.VariableDecl("double", result)]

        if needs_jacobian or needs_inverse_jacobian:
            code += jacobian(L, gdim, tdim, cell_shape)

        if needs_inverse_jacobian:
            code += inverse_jacobian(L, gdim, tdim, cell_shape)
            if tdim != gdim :
                code += orientation(L)

    # Extract variables
    mappings = ir["mappings"]
    offsets = ir["physical_offsets"]

    # Generate bodies for each degree of freedom
    values = L.Symbol("values")
    for (i, dof) in enumerate(ir["dofs"]):
        c, r = _generate_body(L, i, dof, mappings[i], gdim, tdim, cell_shape, offsets[i])
        code += c
        code += [L.Assign(values[i], r)]

    return code



# -*- coding: utf-8 -*-
# Copyright (C) 2017 Garth N. Wells
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from ufl import FiniteElement, VectorElement, MixedElement
from ufl import interval, triangle, tetrahedron
import ffc
import ffc_factory


@pytest.fixture(scope="module")
def point_data():
    """Points are which evaluate functions are tested"""
    p = {1: [(0.114,), (0.349,), (0.986,)],
         2: [(0.114, 0.854), (0.349, 0.247), (0.986, 0.045)],
         3: [(0.114, 0.854, 0.126), (0.349, 0.247, 0.457),
             (0.986, 0.045, 0.127)]}
    return p


@pytest.fixture(scope="module")
def build_ufl_element_list():
    """Build collection of UFL elements"""

    elements = []

    # Lagrange elements
    for cell in (interval, triangle, tetrahedron):
        for p in range(1, 2):
            elements.append(FiniteElement("Lagrange", cell, p))
            elements.append(VectorElement("Lagrange", cell, p))
            elements.append(FiniteElement("Discontinuous Lagrange", cell, p-1))

    # Vector elements
    for cell in (triangle, tetrahedron):
        for p in range(1, 2):
            elements.append(FiniteElement("RT", cell, p))
            elements.append(FiniteElement("BDM", cell, p))
            elements.append(FiniteElement("N1curl", cell, p))
            elements.append(FiniteElement("N2curl", cell, p))
            elements.append(FiniteElement("Discontinuous Raviart-Thomas", cell, p))

    # Mixed elements
    for cell in (interval, triangle, tetrahedron):
        for p in range(1, 3):
            e0 = FiniteElement("Lagrange", cell, p+1)
            e1 = FiniteElement("Lagrange", cell, p)
            e2 = VectorElement("Lagrange", cell, p+1)

            elements.append(MixedElement([e0, e0]))
            elements.append(MixedElement([e0, e1]))
            elements.append(MixedElement([e1, e0]))
            elements.append(MixedElement([e2, e1]))
            elements.append(MixedElement([MixedElement([e1, e1]), e0]))

    for cell in (triangle, tetrahedron):
        for p in range(1, 2):
            e0 = FiniteElement("Lagrange", cell, p+1)
            e1 = FiniteElement("Lagrange", cell, p)
            e2 = VectorElement("Lagrange", cell, p+1)
            e3 = FiniteElement("BDM", cell, p)
            e4 = FiniteElement("RT", cell, p+1)
            e5 = FiniteElement("N1curl", cell, p)

            elements.append(MixedElement([e1, e2]))
            elements.append(MixedElement([e3, MixedElement([e4, e5])]))

    # Misc elements
    for cell in (triangle,):
        for p in range(1, 2):
            elements.append(FiniteElement("HHJ", cell, p))

    for cell in (triangle, tetrahedron):
        elements.append(FiniteElement("CR", cell, 1))
        for p in range(1, 2):
            elements.append(FiniteElement("Regge", cell, p))

    return elements


def element_id(e):
    """Return element string signature as ID"""
    return str(e)


@pytest.fixture(params=build_ufl_element_list(), ids=element_id, scope="module")
def element_pair(request):
    """Given a UFL element, returns UFL element and a JIT-compiled and
    wrapped UFC element.

    """

    ufl_element = request.param
    ufc_element, ufc_dofmap = ffc.jit(ufl_element, parameters=None)
    ufc_element = ffc_factory.make_ufc_finite_element(ufc_element)
    return ufl_element, ufc_element


def test_evaluate_reference_basis_vs_fiat(element_pair, point_data):
    """Tests ufc::finite_element::evaluate_reference_basis against data
    from FIAT.

    """

    ufl_element, ufc_element = element_pair

    # Get geometric and topological dimensions
    tdim = ufc_element.topological_dimension()

    points = point_data[tdim]

    # Create FIAT element via FFC
    fiat_element = ffc.fiatinterface.create_element(ufl_element)

    # Tabulate basis at requsted points via UFC
    values_ufc = ufc_element.evaluate_reference_basis(points)

    # Tabulate basis at requsted points via FIAT
    values_fiat = fiat_element.tabulate(0, points)

    # Extract and reshape FIAT output
    key = (0,)*tdim
    values_fiat = np.array(values_fiat[key])

    # Shape checks
    fiat_value_size = 1
    for i in range(ufc_element.reference_value_rank()):
        fiat_value_size *= values_fiat.shape[i + 1]
        assert values_fiat.shape[i + 1] == ufc_element.reference_value_dimension(i)

    # Reshape FIAT output to compare with UFC output
    values_fiat = values_fiat.reshape((values_fiat.shape[0],
                                       ufc_element.reference_value_size(),
                                       values_fiat.shape[-1]))

    # Transpose
    values_fiat = values_fiat.transpose((2, 0, 1))

    # Check values
    assert np.allclose(values_ufc, values_fiat)

    # FIXME: This could be fragile because it depend on the order of
    # the sub-element in UFL being the same at the UFC order.

    # Test sub-elements recursively
    n = ufc_element.num_sub_elements()
    ufl_subelements = ufl_element.sub_elements()
    assert n == len(ufl_subelements)
    for i, ufl_e in enumerate(ufl_subelements):
        ufc_sub_element = ufc_element.create_sub_element(i)
        test_evaluate_reference_basis_vs_fiat((ufl_e, ufc_sub_element), point_data)


def test_evaluate_basis_vs_fiat(element_pair, point_data):
    """Tests ufc::finite_element::evaluate_basis and
    ufc::finite_element::evaluate_basis_all against data from FIAT.

    """

    ufl_element, ufc_element = element_pair

    # Get geometric and topological dimensions
    gdim = ufc_element.geometric_dimension()

    points = point_data[gdim]

    # Get geometric and topological dimensions
    tdim = ufc_element.topological_dimension()

    # Create FIAT element via FFC
    fiat_element = ffc.fiatinterface.create_element(ufl_element)

    # Tabulate basis at requsted points via FIAT
    values_fiat = fiat_element.tabulate(0, points)

    # Extract and reshape FIAT output
    key = (0,)*tdim
    values_fiat = np.array(values_fiat[key])

    # Shape checks
    fiat_value_size = 1
    for i in range(ufc_element.reference_value_rank()):
        fiat_value_size *= values_fiat.shape[i + 1]
        assert values_fiat.shape[i + 1] == ufc_element.reference_value_dimension(i)

    # Reshape FIAT output to compare with UFC output
    values_fiat = values_fiat.reshape((values_fiat.shape[0],
                                       ufc_element.reference_value_size(),
                                       values_fiat.shape[-1]))

    # Transpose
    values_fiat = values_fiat.transpose((2, 0, 1))

    # Get reference cell
    cell = ufl_element.cell()
    ref_coords = ffc.fiatinterface.reference_cell(cell.cellname()).get_vertices()

    # Iterate over each point and test
    for i, p in enumerate(points):
        values_ufc = ufc_element.evaluate_basis_all(p, ref_coords, 1)
        assert np.allclose(values_ufc, values_fiat[i])

        for d in range(ufc_element.space_dimension()):
            values_ufc = ufc_element.evaluate_basis(d, p, ref_coords, 1)
            assert np.allclose(values_ufc, values_fiat[i][d])

    # Test sub-elements recursively
    n = ufc_element.num_sub_elements()
    ufl_subelements = ufl_element.sub_elements()
    assert n == len(ufl_subelements)
    for i, ufl_e in enumerate(ufl_subelements):
        ufc_sub_element = ufc_element.create_sub_element(i)
        test_evaluate_basis_vs_fiat((ufl_e, ufc_sub_element), point_data)


@pytest.mark.parametrize("order", range(4))
def test_evaluate_reference_basis_deriv_vs_fiat(order, element_pair,
                                                point_data):
    """Tests ufc::finite_element::evaluate_reference_basis_derivatives
    against data from FIAT.

    """

    ufl_element, ufc_element = element_pair

    # Get geometric and topological dimensions
    tdim = ufc_element.geometric_dimension()
    gdim = ufc_element.topological_dimension()

    points = point_data[tdim]

    # Create FIAT element via FFC
    fiat_element = ffc.fiatinterface.create_element(ufl_element)

    # Tabulate basis at requsted points via UFC
    values_ufc = ufc_element.evaluate_reference_basis_derivatives(order, points)

    # Tabulate basis at requsted points via FIAT
    values_fiat = fiat_element.tabulate(order, points)

    # Compute 'FFC' style derivative indicators
    deriv_order = order
    geo_dim = tdim
    num_derivatives = geo_dim**deriv_order
    combinations = [[0] * deriv_order for n in range(num_derivatives)]
    for i in range(1, num_derivatives):
        for k in range(i):
            j = deriv_order - 1
            while j + 1 > 0:
                j -= 1
                if combinations[i][j] + 1 > geo_dim - 1:
                    combinations[i][j] = 0
                else:
                    combinations[i][j] += 1
                    break

    def to_fiat_tuple(comb, gdim):
        """Convert a list of combinations of derivatives to a fiat tuple of
        derivatives.  FIAT expects a list with the number of
        derivatives in each spatial direction.  E.g., in 2D: u_{xyy}
        --> [0, 1, 1] in FFC --> (1, 2) in FIAT.

        """
        new_comb = [0] * gdim
        if comb == []:
            return tuple(new_comb)
        for i in range(gdim):
            new_comb[i] = comb.count(i)
        return tuple(new_comb)

    derivs = [to_fiat_tuple(c, tdim) for c in combinations]
    assert len(derivs) == tdim**order

    # Iterate over derivatives
    for i, d in enumerate(derivs):
        values_fiat_slice = np.array(values_fiat[d])

        # Shape checks
        fiat_value_size = 1
        for j in range(ufc_element.reference_value_rank()):
            fiat_value_size *= values_fiat_slice.shape[j + 1]
            assert values_fiat_slice.shape[j + 1] == ufc_element.reference_value_dimension(j)

        # Reshape FIAT output to compare with UFC output
        values_fiat_slice = values_fiat_slice.reshape((values_fiat_slice.shape[0],
                                                       ufc_element.reference_value_size(),
                                                       values_fiat_slice.shape[-1]))

        # Transpose
        values_fiat_slice = values_fiat_slice.transpose((2, 0, 1))

        # Compare
        assert np.allclose(values_ufc[:,:,i,:], values_fiat_slice)

    # Test sub-elements recursively
    n = ufc_element.num_sub_elements()
    ufl_subelements = ufl_element.sub_elements()
    assert n == len(ufl_subelements)
    for i, ufl_e in enumerate(ufl_subelements):
        ufc_sub_element = ufc_element.create_sub_element(i)
        test_evaluate_reference_basis_deriv_vs_fiat(order, (ufl_e, ufc_sub_element), point_data)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2017 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

from ffc.uflacs.backends.ufc.generator import ufc_generator, ufc_integral_types
from ufl.utils.formatting import dstr


# (This is currently not in use, see ffc/codegeneration.py for current status)
class ufc_integral(ufc_generator):
    "Each function maps to a keyword in the template. See documentation of ufc_generator."
    def __init__(self, integral_type):
        assert integral_type in ufc_integral_types
        ufc_generator.__init__(self, integral_type + "_integral")

    def enabled_coefficients(self, L, ir):
        enabled_coefficients = ir["enabled_coefficients"]
        initializer_list = ", ".join("true" if enabled else "false"
                                     for enabled in enabled_coefficients)
        code = L.StatementList([
            # Cheating a bit with verbatim:
            L.VerbatimStatement("static const std::vector<bool> enabled({%s});" % initializer_list),
            L.Return(L.Symbol("enabled")),
            ])
        return code

    def tabulate_tensor(self, L, ir):
        # FIXME: This is where the current ffc backend code generation should be injected,
        #        however that will require using pick_representation in here?
        tt = ir["tabulate_tensor"]
        code = "code generated from %s" % tt
        return code

    def tabulate_tensor_comment(self, L, ir):
        "Generate comment for tabulate_tensor."

        r = ir["representation"]
        integrals_metadata = ir["integrals_metadata"]
        integral_metadata = ir["integral_metadata"]

        lines  = [
            "This function was generated using '%s' representation" % r,
            "with the following integrals metadata:",
            ""]
        lines += [("  " + l) for l in dstr(integrals_metadata).split("\n")]
        lines += [""]
        for i, metadata in enumerate(integral_metadata):
            lines += [
                "",
                "and the following integral %d metadata:" % i,
                "",
                ]
            lines += [("  " + l) for l in dstr(metadata).split("\n")][:-1]
            lines += [""]

        return [L.Comment(line) for line in lines]


class ufc_cell_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "cell")

class ufc_exterior_facet_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "exterior_facet")

class ufc_interior_facet_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "interior_facet")

class ufc_vertex_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "vertex")

class ufc_custom_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "custom")

    def num_cells(self, L, ir):
        value = ir["num_cells"]
        return L.Return(L.LiteralInt(value))

class ufc_interface_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "interface")

class ufc_overlap_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "overlap")

class ufc_cutcell_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "cutcell")



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2017 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

# Note: Most of the code in this file is a direct translation from the old implementation in FFC

from ffc.classname import make_integral_classname
from ffc.uflacs.backends.ufc.generator import ufc_generator, integral_name_templates, ufc_integral_types
from ffc.uflacs.backends.ufc.utils import generate_return_new, generate_return_new_switch


def create_delegate(integral_type, declname, impl):
    def _delegate(self, L, ir, parameters):
        return impl(self, L, ir, parameters, integral_type, declname)
    _delegate.__doc__ = impl.__doc__ % {"declname": declname, "integral_type": integral_type}
    return _delegate


def add_ufc_form_integral_methods(cls):
    """This function generates methods on the class it decorates,
    for each integral name template and for each integral type.

    This allows implementing e.g. create_###_integrals once in the
    decorated class as '_create_foo_integrals', and this function will
    expand that implementation into 'create_cell_integrals',
    'create_exterior_facet_integrals', etc.

    Name templates are taken from 'integral_name_templates' and 'ufc_integral_types'.
    """
    # The dummy name "foo" is chosen for familiarity for ffc developers
    dummy_integral_type = "foo"

    for template in integral_name_templates:
        implname = "_" + (template % (dummy_integral_type,))
        impl = getattr(cls, implname)
        for integral_type in ufc_integral_types:
            declname = template % (integral_type,)
            _delegate = create_delegate(integral_type, declname, impl)
            setattr(cls, declname, _delegate)
    return cls


@add_ufc_form_integral_methods
class ufc_form(ufc_generator):
    """Each function maps to a keyword in the template. See documentation of ufc_generator.

    The exceptions are functions on the form
        def _*_foo_*(self, L, ir, parameters, integral_type, declname)
    which add_ufc_form_integral_methods will duplicate for foo = each integral type.
    """
    def __init__(self):
        ufc_generator.__init__(self, "form")

    def num_coefficients(self, L, num_coefficients):
        return L.Return(num_coefficients)

    def rank(self, L, rank):
        return L.Return(rank)

    def original_coefficient_position(self, L, ir):
        i = L.Symbol("i")
        positions = ir["original_coefficient_position"]

        # Check argument
        msg = "Invalid original coefficient index."

        if positions:
            code = [
            L.If(L.GE(i, len(positions)),
                L.Throw("std::runtime_error", msg))
            ]

            # Throwing a lot into the 'typename' string here but
            # no plans for building a full C++ type system
            typename = "static const std::vector<std::size_t>"
            position = L.Symbol("position")
            initializer_list = L.VerbatimExpr("{" + ", ".join(str(i) for i in positions) + "}")
            code += [
                L.VariableDecl(typename, position, value=initializer_list),
                L.Return(position[i]),
                ]
            return code
        else:
            code = [ L.Throw("std::runtime_error", msg),
                     L.Return(i) ]
        return code

    def create_coordinate_finite_element(self, L, ir):
        classnames = ir["create_coordinate_finite_element"]
        assert len(classnames) == 1
        return generate_return_new(L, classnames[0], factory=ir["jit"])

    def create_coordinate_dofmap(self, L, ir):
        classnames = ir["create_coordinate_dofmap"]
        assert len(classnames) == 1
        return generate_return_new(L, classnames[0], factory=ir["jit"])

    def create_coordinate_mapping(self, L, ir):
        classnames = ir["create_coordinate_mapping"]
        assert len(classnames) == 1  # list of length 1 until we support multiple domains
        return generate_return_new(L, classnames[0], factory=ir["jit"])

    def create_finite_element(self, L, ir):
        i = L.Symbol("i")
        classnames = ir["create_finite_element"]
        return generate_return_new_switch(L, i, classnames, factory=ir["jit"])

    def create_dofmap(self, L, ir):
        i = L.Symbol("i")
        classnames = ir["create_dofmap"]
        return generate_return_new_switch(L, i, classnames, factory=ir["jit"])


    # This group of functions are repeated for each
    # foo_integral by add_ufc_form_integral_methods:

    def _max_foo_subdomain_id(self, L, ir, parameters, integral_type, declname):
        "Return implementation of ufc::form::%(declname)s()."
        # e.g. max_subdomain_id = ir["max_cell_subdomain_id"]
        max_subdomain_id = ir[declname]
        return L.Return(int(max_subdomain_id))

    def _has_foo_integrals(self, L, ir, parameters, integral_type, declname):
        "Return implementation of ufc::form::%(declname)s()."
        # e.g. has_integrals = ir["has_cell_integrals"]
        has_integrals = ir[declname]
        return L.Return(bool(has_integrals))

    def _create_foo_integral(self, L, ir, parameters, integral_type, declname):
        "Return implementation of ufc::form::%(declname)s()."
        # e.g. subdomain_ids, classnames = ir["create_cell_integral"]
        subdomain_ids, classnames = ir[declname]
        subdomain_id = L.Symbol("subdomain_id")
        return generate_return_new_switch(L, subdomain_id, classnames,
                                          subdomain_ids, factory=ir["jit"])

    def _create_default_foo_integral(self, L, ir, parameters, integral_type, declname):
        "Return implementation of ufc::form::%(declname)s()."
        # e.g. classname = ir["create_default_cell_integral"]
        classname = ir[declname]
        if classname is None:
            return L.Return(L.Null())
        else:
            return generate_return_new(L, classname, factory=ir["jit"])



from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
import sys
import setuptools
import ffc

__version__ = '0.0.1'


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


ext_modules = [
    Extension('ffc_factory',
              ['interface.cpp',
               'dofmap.cpp',
               'finite_element.cpp',
               'finite_element_bindings.cpp'],
              include_dirs=[
                  # Path to pybind11 headers
                  ffc.get_include_path(),
                  get_pybind_include(),
                  get_pybind_include(user=True)
              ],
              language='c++'),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=default'):
                opts.append('-fvisibility=default')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


setup(name='fenics-ffc-factory',
      version=__version__,
      author='FEniCS Project',
      author_email='fenics-dev@googlegroups.com',
      url='https://fenicsproject.org',
      description='Python wrappers for UFC using pybind11',
      long_description='',
      packages=find_packages(),
      ext_modules=ext_modules,
      setup_requires=['pybind11>=1.7',
                      'fenics-ffc'],
      cmdclass={'build_ext': BuildExt},
      zip_safe=False,)



# -*- coding: utf-8 -*-
"This module defines some basics for generating C++ code."

# Copyright (C) 2009-2017 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard 2011
# Modified by Marie E. Rognes 2010
# Modified by Martin Sandve Alnæs 2013-2017
# Modified by Chris Richardson 2017

# ufc class names

def make_classname(prefix, basename, signature):
    pre = prefix.lower() + "_" if prefix else ""
    sig = str(signature).lower()
    return "%s%s_%s" % (pre, basename, sig)


def make_integral_classname(prefix, integral_type, form_id, subdomain_id):
    basename = "%s_integral_%s" % (integral_type, str(form_id).lower())
    return make_classname(prefix, basename, subdomain_id)



def optimize_integral_ir(ir, parameters):
    ir = ir.copy()
    integral_data, form_data, prefix, parameters = ir["compile_integral"]
    parameters = parameters.copy()
    parameters.setdefault("mode", "spectral")  # default optimization mode
    ir["compile_integral"] = (integral_data, form_data, prefix, parameters)
    return ir



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2017 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

from six import string_types

# TODO: Move these to uflacs.language utils?


def generate_return_new(L, classname, factory):
    if factory:
        return L.Return(L.Call("create_" + classname))
    else:
        return L.Return(L.New(classname))


def generate_return_new_switch(L, i, classnames, args=None, factory=False):
    # TODO: UFC functions of this type could be replaced with return vector<shared_ptr<T>>{objects}.

    if isinstance(i, string_types):
        i = L.Symbol(i)

    if factory:
        def create(classname):
            return L.Call("create_" + classname)
    else:
        def create(classname):
            return L.New(classname)

    default = L.Return(L.Null())
    if classnames:
        cases = []
        if args is None:
            args = list(range(len(classnames)))
        for j, classname in zip(args, classnames):
            if classname:
                cases.append((j, L.Return(create(classname))))
        return L.Switch(i, cases, default=default)
    else:
        return default


def generate_return_literal_switch(L, i, values, default, literal_type, typename=None):
    # TODO: UFC functions of this type could be replaced with return vector<T>{values}.

    if isinstance(i, string_types):
        i = L.Symbol(i)
    return_default = L.Return(literal_type(default))

    if values and typename is not None:
        # Store values in static table and return from there
        V = L.Symbol("return_values")
        decl = L.ArrayDecl("static const %s" % typename, V, len(values),
                            [literal_type(k) for k in values])
        return L.StatementList([
            decl,
            L.If(L.GE(i, len(values)),
                 return_default),
            L.Return(V[i])
            ])
    elif values:
        # Need typename to create static array, fallback to switch
        cases = [(j, L.Return(literal_type(k)))
                 for j, k in enumerate(values)]
        return L.Switch(i, cases, default=return_default)
    else:
        # No values, just return default
        return return_default


def generate_return_sizet_switch(L, i, values, default):
    return generate_return_literal_switch(L, i, values, default, L.LiteralInt, "std::size_t")


def generate_return_int_switch(L, i, values, default):
    return generate_return_literal_switch(L, i, values, default, L.LiteralInt, "int")


def generate_return_bool_switch(L, i, values, default):
    return generate_return_literal_switch(L, i, values, default, L.LiteralBool, "bool")


# TODO: Better error handling
def generate_error(L, msg, emit_warning):
    if emit_warning:
        return L.VerbatimStatement('std::cerr << "*** FFC warning: " << "%s" << std::endl;' % (msg,))
    else:
        return L.Throw('std::runtime_error', msg)


"""
// TODO: UFC functions that return constant arrays
// could use something like this to reduce copying,
// returning pointers to static data:

  template<T>
  class array_view
  {
  public:
    array_view(std::size_t size, const T * data):
      size(size), data(data)
    const std::size_t size;
    const T * data;
    const T operator[](std::size_t i) const
    { return data[i]; }
  };

  array_view<int> form::original_coefficient_positions() const
  {
    static const int data = { 0, 1, 2, 5 };
    return array_view<int>{4, data};
  }

  array_view<bool> integral::enabled_coefficients() const
  {
    static const bool data = { true, true, true, false, false, true };
    return array_view<bool>{6, data};
  }
"""
